# Copyright (C) 2018 VMware, Inc. All rights reserved.

import subprocess


def _check_output(cmd):
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    out, err = process.communicate()
    return out


def find_bridge(name):
    cmd = ["ovs-vsctl", "list-br"]
    return name in _check_output(cmd).split()


def delete_bridge(name):
    if not find_bridge(name):
        return
    cmd = ["ovs-vsctl", "--timeout=5", "del-br", name]
    subprocess.call(cmd)


def create_bridge(name, external_id=None, mode="secure", mac=None):
    if find_bridge(name):
        return
    cmd = ["ovs-vsctl", "--timeout=5", "add-br", name, "--",
           "set", "bridge", name,
           "external_ids:bridge-id=%s" % (external_id or name),
           "fail_mode=%s" % mode]
    if mac:
        cmd.append("other_config:hwaddr=%s" % mac)
    print("cmd is %s" % " ".join(cmd))
    subprocess.call(cmd)


def find_port(name, br_name):
    cmd = ["ovs-vsctl", "list-ports", br_name]
    return name in _check_output(cmd).split()


def create_port(name, br_name):
    assert find_bridge(br_name), "bridge %s not found" % br_name
    if find_port(name, br_name):
        return
    cmd = ["ovs-vsctl", "--timeout=5", "add-port", br_name, name]
    subprocess.call(cmd)

def add_ip_to_bridge(br_name, ip):
    assert find_bridge(br_name), "bridge %s not found" % br_name
    cmd = ["ifconfig", br_name, ip]
    subprocess.call(cmd)

def verify_topo(bridges, vteps):
    # verify ovs is running
    ovs_status = _check_output(["/usr/share/openvswitch/scripts/ovs-ctl", "status"])
    assert ovs_status.find("is running") >= 0, ovs_status

    # for each vtep interface, verify
    # it exists and it does not has IP
    for vtep in vteps:
        print("vtep is %s" % vtep)
        if_status = _check_output(["ifconfig", vtep])
        assert isinstance(if_status, str) and if_status.find(vtep) == 0, (
            "cannot find interface %s" % vtep)


def get_intf_mac(intf):
    cmd = ["ifconfig", intf]
    out = _check_output(cmd).split()
    mac_key_word = ["HWaddr", "ether"]
    for key in mac_key_word:
        if key in out:
            idx = out.index(key)
            break
    return out[idx + 1]


def create_topo(bridges, vteps, ips=[]):
    verify_topo(bridges, vteps)
    t_br = bridges[-1]
    for bridge in bridges:
        if bridge is t_br:
            break
        create_bridge(bridge[0], mode=bridge[1])
    create_bridge(t_br[0], mode=t_br[1], mac=get_intf_mac(vteps[0]))
    for vtep in vteps:
        create_port(vtep, t_br[0])
    for i in range(len(ips)):
        add_ip_to_bridge(t_br[0], ips[i])


def clear_topo():
    cmd = ["ovs-vsctl", "list-br"]
    bridges = _check_output(cmd).split()
    for bridge in bridges:
        delete_bridge(bridge)

def get_ovsdb_manager(name):
    cmd = ["ovs-vsctl", "get-manager"]
    return name in _check_output(cmd).split()

def set_ovsdb_manager():
    nsxagent_manager = "unix:/var/run/vmware/nsx-agent/nsxagent_ovsdb.sock"
    if get_ovsdb_manager(nsxagent_manager):
        return
    cmd = ["ovs-vsctl", "--timeout=5", "set-manager", nsxagent_manager]
    subprocess.call(cmd)

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(prog="__file__")
    parser.add_argument("-i", "--br_int", nargs="?", default="nsx-managed",
                        help="name of integration bridge")
    parser.add_argument("-t", "--br_trans", nargs="?", default="transport",
                        help="name of transport bridge")
    parser.add_argument("-e", "--vtep", nargs="?", default="eth1",
                        help="name of vtep device")
    parser.add_argument("-a", "--ip_address", nargs="?", default="",
                        help="name of vtep device")
    parser.add_argument("-c", "--clear", action="store_true",
                        help="clear all bridges setups")
    parser.add_argument("-n", "--no_trans", action="store_true",
                        help="do not create transport bridge")

    set_ovsdb_manager()

    args = parser.parse_args()
    if args.no_trans:
        create_bridge(args.br_int, mode="secure")
    elif not args.clear:
        bridges = []
        bridges.append([args.br_int, "secure"])
        bridges.append([args.br_trans, "standalone"])
        vteps = [args.vtep]
        if args.ip_address:
            create_topo(bridges, vteps, [args.ip_address])
        else:
            create_topo(bridges, vteps)
    else:
        clear_topo()
