# ***************************************************************************
# Copyright 2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import json
import nsx_logging
import cli.commands.host_shared.logical_switch_util as ls_util
from cli.commands.lcp_common.cfgagent.client import get_cfgagent_cli_output
from cli.utils.error_code import ErrorCode as ec

lg = nsx_logging.getLogger("cli.commands.lcp_common.cfgagent.logical_switch_cfgagent_util")


# There is a separate API to get VNI in CFGAGENT, not need to get full LS info
def get_switch_vni(uuid_val):
    lg.debug("Get logical switch VNI from CFGAGENT, uuid %s", uuid_val)
    response = get_cfgagent_cli_output("cfgagent/statecache/logicalswitchvni", [uuid_val])
    return parse_logical_switch_vni(response)


def parse_logical_switch_vni(vni_resp):
    try:
        vni_resp_decoded = json.loads(vni_resp)
    except (TypeError, ValueError) as err:
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        return
    return vni_resp_decoded.get("vni")


def get_switches():
    lg.debug("Get logical switches from CFGAGENT")
    response = get_cfgagent_cli_output("cfgagent/statecache/logicalswitches")
    return parse_logical_switches(response)


def parse_logical_switches(lswitches_resp):
    try:
        lswitches_resp_decoded = json.loads(lswitches_resp)
    except (TypeError, ValueError) as err:
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        return
    ls_list = [parse_logical_switch_entry(ls_data) for ls_data in lswitches_resp_decoded]
    return ls_list


def get_switch_by_uuid(uuid_val):
    lg.debug("Get logical switch from CFGAGENT, uuid %s", uuid_val)
    response = get_cfgagent_cli_output("cfgagent/statecache/logicalswitch", ['uuid', uuid_val])
    return parse_logical_switch(response)


def get_switch_by_vni(vni_val):
    lg.debug("Get logical switch from CFGAGENT, vni %s", vni_val)
    response = get_cfgagent_cli_output("cfgagent/statecache/logicalswitch", ['vni', vni_val])
    return parse_logical_switch(response)


def get_switch_by_vlan(vlan_val):
    lg.debug("Get logical switch from CFGAGENT, vlan %s", vlan_val)
    response = get_cfgagent_cli_output("cfgagent/statecache/logicalswitch", ['vlan', vlan_val])
    return parse_logical_switch(response)


def parse_logical_switch(ls_resp):
    try:
        ls_resp_decoded = json.loads(ls_resp)
    except (TypeError, ValueError) as err:
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        return
    return parse_logical_switch_entry(ls_resp_decoded)


def parse_logical_switch_entry(ls_data):
    ls_entry = ls_util.LogicalSwitch()
    ls_entry.uuid_val = ls_data.get("uuid")
    ls_entry.vni = ls_data.get("vni")
    ls_entry.name = ls_data.get("switch_name")
    ls_entry.replication_mode = ls_data.get("replication_mode")
    ls_entry.transport_binding = ls_data.get("transport_binding")
    ls_entry.vlan_id = ls_data.get("vlan_id")
    ls_entry.admin_state_up = ls_data.get("ls_admin_state_up")

    return ls_entry


def get_mac_table(uuid_val):
    lg.debug("Get MAC table from CFGAGENT, uuid %s", uuid_val)
    response = get_cfgagent_cli_output("cfgagent/statecache/L2mac", [uuid_val])
    return parse_mac_table(response)


def parse_mac_table(mac_resp):
    try:
        mac_resp_decoded = json.loads(mac_resp)
    except (TypeError, ValueError) as err:
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        return

    remote_mac_list = mac_resp_decoded.get("remote_cache")
    mac_table_remote = parse_mac_list(remote_mac_list)

    local_mac_list = mac_resp_decoded.get("local_cache")
    mac_table_local = parse_mac_list(local_mac_list)

    return (mac_table_remote, mac_table_local)


def parse_mac_list(mac_list):
    mac_table = []
    for mac_data in mac_list or []:
        mac_entry = ls_util.MacTableEntry()
        mac_entry.inner_mac = mac_data.get("vm_mac")
        mac_entry.outer_mac = mac_data.get("vtep_mac")
        mac_entry.outer_ip = mac_data.get("vtep_ip")
        mac_table.append(mac_entry)
    return mac_table


def get_vtep_table(uuid_val, getHwVTEPs):
    lg.debug("Get VTEP table from CFGAGENT, uuid %s, getHwVTEPs %d",
             uuid_val, getHwVTEPs)
    response = get_cfgagent_cli_output("cfgagent/statecache/L2vtep",
                                       [uuid_val, getHwVTEPs])
    return parse_vtep_table(response)


def parse_vtep_table(vtep_resp):
    try:
        vtep_resp_decoded = json.loads(vtep_resp)
    except (TypeError, ValueError) as err:
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        return

    remote_vtep_list = vtep_resp_decoded.get("remote_cache")
    vtep_table_remote = parse_vtep_list(remote_vtep_list)

    local_vtep_list = vtep_resp_decoded.get("local_cache")
    vtep_table_local = parse_vtep_list(local_vtep_list)

    return (vtep_table_remote, vtep_table_local)


def parse_vtep_list(vtep_list):
    vtep_table = []
    for vtep_data in vtep_list or []:
        vtep_entry = ls_util.VtepTableEntry()
        vtep_entry.ip_addr = vtep_data.get("vtep_ip")
        vtep_entry.seg_id = vtep_data.get("segment_id")
        vtep_entry.label = vtep_data.get("vtep_label")
        vtep_entry.mac = vtep_data.get("vtep_mac")
        vtep_entry.dev_name = vtep_data.get("dev_name")
        vtep_table.append(vtep_entry)
    return vtep_table


def get_arp_table(uuid_val):
    lg.debug("Get ARP table from CFGAGENT, uuid %s", uuid_val)
    response = get_cfgagent_cli_output("cfgagent/statecache/L2arp", [uuid_val])
    return parse_arp_table(response)


def parse_arp_table(arp_resp):
    try:
        arp_resp_decoded = json.loads(arp_resp)
    except (TypeError, ValueError) as err:
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        return

    remote_arp_list = arp_resp_decoded.get("remote_cache")
    arp_table_remote = parse_arp_list(remote_arp_list)

    local_arp_list = arp_resp_decoded.get("local_cache")
    arp_table_local = parse_arp_list(local_arp_list)

    return (arp_table_remote, arp_table_local)


def parse_arp_list(arp_list):
    arp_table = []
    for arp_data in arp_list or []:
        arp_entry = ls_util.ArpTableEntry()
        arp_entry.ip = arp_data.get("ip")
        arp_entry.mac = arp_data.get("mac")
        arp_table.append(arp_entry)
    return arp_table


def get_nd_table(uuid_val):
    lg.debug("Get ND table from CFGAGENT, uuid %s", uuid_val)
    response = get_cfgagent_cli_output("cfgagent/statecache/L2ND", [uuid_val])
    return parse_nd_table(response)


def parse_nd_table(nd_resp):
    try:
        nd_resp_decoded = json.loads(nd_resp)
    except (TypeError, ValueError) as err:
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        return

    remote_nd_list = nd_resp_decoded.get("remote_cache")
    nd_table_remote = parse_nd_list(remote_nd_list)

    local_nd_list = nd_resp_decoded.get("local_cache")
    nd_table_local = parse_nd_list(local_nd_list)

    return (nd_table_remote, nd_table_local)


def parse_nd_list(nd_list):
    nd_table = []
    for nd_data in nd_list or []:
        nd_entry = ls_util.NDTableEntry()
        nd_entry.ip = nd_data.get("ip")
        nd_entry.mac = nd_data.get("mac")
        nd_table.append(nd_entry)
    return nd_table


def get_dr_vtep():
    response = get_cfgagent_cli_output("cfgagent/statecache/dr-vtep")
    return parse_dr_vtep(response)


def get_routing_domain_vteps():
    response = get_cfgagent_cli_output("cfgagent/statecache/routing-domains")
    return parse_routing_domain_vteps(response)


def get_routing_domain_vteps_uuid(rd_uuid):
    response = get_cfgagent_cli_output("cfgagent/statecache/routing-domain-uuid", [rd_uuid])
    return parse_routing_domain_vteps(response)


def parse_dr_vtep(response):
    try:
       result = json.loads(response)
    except (TypeError, ValueError) as err:
       lg.error(err, error_code=ec.INTERNAL_ERROR)
       return
    return result


def parse_routing_domain_vteps(response):
    try:
       result = json.loads(response)
    except (TypeError, ValueError) as err:
       lg.error(err, error_code=ec.INTERNAL_ERROR)
       return
    return result
