# ***************************************************************************
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************


__all__ = ["CliCfgSerializerDescriptor"]


import nsx_logging
import os
import re
try:
    import simplejson
except ImportError:
    import json as simplejson
import yaml

from lib.json_schema import JsonSchema
from cli.descriptors.descriptor import Descriptor
from cli.descriptors.descriptor import SCHEMA_REGISTRY
from cli.utils.import_utils import get_class
from cli.utils.error_code import ErrorCode as ec


lg = nsx_logging.getLogger("cli.descriptors.cli_cfg_serializer_descriptor")


CLI_BASE_DIR = os.path.dirname(os.path.dirname(__file__))
CLI_SERIALIZER_SCHEMA_FILE = os.path.join(os.path.dirname(__file__),
                                       'cli_cfg_serializer_descriptor.yml')
CLI_SERIALIZER_TYPES = (dict((r.id, r)
                        for r in yaml.load(open(CLI_SERIALIZER_SCHEMA_FILE), Loader=yaml.Loader)))
for key in list(CLI_SERIALIZER_TYPES.keys()):
    if key not in SCHEMA_REGISTRY:
        JsonSchema(CLI_SERIALIZER_TYPES[key].raw_schema,
                   schema_registry=SCHEMA_REGISTRY, type_path=[key],
                   strict=True)


class CliCfgSerializerDescriptor(Descriptor):
    """Descriptor specification of a CLI configuration serializer.

    The CLI supports the command "show configuration [category]" which
    shows the current configured values of all user editable configuration
    for a given optional category.  What's shown for a given configurable
    value is the command a user would use to configure the value (e.g.
    "set hostname testserver").

    All user configurable values need to declare a CliCfgSerializerDescriptor
    that specifies a class which implements a __call__ method that returns
    the configuration value in command form.
    """
    yaml_tag = '!CliCfgSerializer'

    def __init__(self, id, schema_spec):
        super(CliCfgSerializerDescriptor, self).__init__(id, schema_spec)
        self._factory = None  # Set by factory
        self._raw_schema = schema_spec
        self._name = schema_spec["name"]
        self._module = schema_spec.get("module")
        self._class_name = schema_spec.get("class_name")

    @property
    def name(self):
        return self._name

    def get_serialize_output(self):
        """Return command form of a configuration value.

        Return:
            The command form of a configuration value.
        """
        try:
            cls = get_class(self._module, self._class_name)
        except ImportError:
            msg = ("Error resolving query argument '%s' module '%s' or "
                   "class_name '%s'" %
                   (self._name, self._module, self._class_name))
            lg.exception(msg, error_code=ec.INVALID_ARGUMENT)
            assert False, msg
            return []
        result = cls(self)()
        return result
