# **********************************************************************
# Copyright 2017 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import nsx_logging
import threading

from cli.utils import rest_utils
from compat23 import compat23_httplib


lg = nsx_logging.getLogger(__name__)


_NODE_TYPE_LABEL_CONTROLLER = "ctl"
_NODE_TYPE_LABEL = {
    "ControllerClusterRoleConfig": _NODE_TYPE_LABEL_CONTROLLER,
    "EdgeNode": "edg",
    "ESXI": "esx",
    "ManagementClusterRoleConfig": "mgr",
    "PublicCloudGatewayNode": "pcg",
    "CENTOSKVM": "kvm",
    "SLESKVM": "kvm",
    "RHELKVM": "kvm",
    "UBUNTUKVM": "kvm",
    "UBUNTUSERVER": "srv",
    "RHELSERVER": "srv",
    "CENTOSSERVER": "srv",
    "SLESSERVER": "srv",
    "RHELCONTAINER": "srv",
    "CENTOSCONTAINER": "srv",
    "WINDOWSSERVER": "srv",
    "OELSERVER": "srv",
}
_NODE_TYPE_LABEL_UNKNOWN = "???"


def get_nodes():
    """Get dict of nodes, with display_uuid as key, and a dict of uuid,
    display_name, node_type as values.

    {
      '0c77a7c0-0b3e-11e7-b7a3-970a64d76956':
      {
          'uuid': 0C77A7C0-0B3E-11E7-B7A3-970A64D76956,
          'display_name': host1,
          'node_type': 'ESXI'
      }, ...
    }
    """
    results = {}
    all_nodes = {} # Dictionary of cluster and fabric nodes

    for t in ('fabric', 'cluster'):
        status, json, err, _ = rest_utils.local_node_request(
            'GET', '/%s/nodes' % t, proton_request=True, retry_time=1)
        if status == compat23_httplib.OK:
            all_nodes[t] = json['results']

    # For unified appliance/converged appliance, keep track of all the manager
    # IP addresses. This would allow to not add any role, which is running on
    # same machine as manager to be exposed as standalone node.
    manager_ips = [node.get('appliance_mgmt_listen_addr')
                   for node in all_nodes.get('cluster', [])
                   if 'appliance_mgmt_listen_addr' in node
                  ]

    for nodes in list(all_nodes.values()):
        for node in nodes:
            uuid = node['id']
            display_uuid = uuid.lower()  # Keep displayed uuid lowercase
            node_data = prepare_node_data(node)
            # Ignore controllers now that it's a role in the unified appliance
            if node_data.get('node_type_label') == _NODE_TYPE_LABEL_CONTROLLER:
                continue
            # Only add if not sharing the IP with the manager
            if (node_data.pop('ip', None) not in manager_ips or
                node_data.get('node_type_label') == 'mgr'):
                results[display_uuid] = node_data

    return results


def prepare_node_data(node):
    """Massage the node data received from API to more consumable format"""

    uuid = node['id']
    display_name = node['display_name']
    node_type = node['resource_type']
    ip_addr = None # Only relevant for Manager and Controller

    type_override = node_type

    if node_type == 'ClusterNodeConfig':
        if 'manager_role' in node:
            type_override = node.get('manager_role', {}).get('type')
            ip_addr = node.get('appliance_mgmt_listen_addr')
        elif 'controller_role' in node:
            type_override = node.get('controller_role', {}).get('type')
            ip_addr = node.get('controller_role', {}
                              ).get('control_plane_listen_addr', {}
                              ).get('ip_address')
    elif node_type == 'HostNode':
        type_override = node.get('os_type')

    result = {'uuid': uuid,
              'display_name': display_name,
              'ip': ip_addr,
              'node_type': type_override,
              'node_type_label': _NODE_TYPE_LABEL.get(
              type_override, _NODE_TYPE_LABEL_UNKNOWN)
             }
    return result


# TODO Use DCN (Data Change Notification) to avoid having to keep such thread
class CentralCLICacheThread(threading.Thread):
    """This will cache API responses to improve responsiveness of NSX-CLI

    Currently it only caches fabric and cluster nodes, once every 10 seconds.
    But in future, it can be further extended to cache more values.
    """
    def __init__(self, descriptor_factory, service, stop_event, delay=10):
        threading.Thread.__init__(self)
        self.descriptor_factory = descriptor_factory
        self.service = service
        self.stop_event = stop_event
        self.delay = delay

    def run(self):
        # Populate node_cache first, and then populate Central CLI descriptors
        self.descriptor_factory.node_cache = get_nodes()
        self.descriptor_factory.process_central_cli_descriptors(self.service)

        # Refresh node_cache periodically, to account for change in cluster
        # membership
        while not self.stop_event.is_set():
            self.descriptor_factory.node_cache = get_nodes()
            self.stop_event.wait(self.delay)


class VerifyArgCache(object):
    """Cache for arguments of type verify arg.

    This will keep result of possible values for a verify argument, until
    the cache gets cleared. This is being used to optimize CLI run-time
    performance, as it avoids recalculating possible values of verify arg
    type, for given command list.
    """
    def __init__(self):
        self.clear()

    def clear(self):
        self._verify_arg_cache = {}
        self._verify_arg_req_dict = None

    def set_arg_values(self, req_dict, arg_id, arg_values):
        if req_dict != self._verify_arg_req_dict:
             self.clear()
             self._verify_arg_req_dict = req_dict
        self._verify_arg_cache[arg_id] = arg_values

    def get_arg_values(self, req_dict, arg_id):
        if req_dict != self._verify_arg_req_dict:
            self.clear()
            return None
        else:
            return self._verify_arg_cache.get(arg_id)
