# **********************************************************************
# Copyright 2017 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import nsx_logging
import uuid
import signal
import time

from cli.utils import rest_utils
from compat23 import compat23_httplib

lg = nsx_logging.getLogger(__name__)

BASE_URL = '/platform-management/central-cli?action=%s'

ASYNC_POOL_SIZE = 5
DELAY_BETWEEN_QUERY_REQUESTS = 2  # Number of seconds between status queries


class CentralCLIAPIException(Exception):
    pass


def invoke_command(node, cmd, modifier, context,
                   username, rbac_permissions, debug_mode):
    '''Invoke a remote command

    Return the unique request id for command queued for execution.
    '''
    request_id = str(uuid.uuid4())  # Random and unique request id
    path = BASE_URL % 'invoke_command'
    modifier = None if not modifier else modifier
    # Import pwd here instead of globally as pwd may not be available on all
    # platforms and is only needed on MP for processing Central CLI requests
    import pwd
    status, json, text, resp = rest_utils.local_node_request(
        "POST",
        path=path,
        # TODO: to workaround PR 2266424, use the username field
        # to store a string representation of username's UID. In
        # a future release we should add a "userid" property and
        # make the corresponding changes in protobuf, mp, and the
        # platform-client
        body={
            "node_uuid": node,
            "request_id": request_id,
            "command": cmd,
            "modifier": modifier,
            "context": context,
            "username": str(pwd.getpwnam(username).pw_uid),
            "rbac_permissions": rbac_permissions,
            "debug_mode": debug_mode
        },
        proton_request=True)

    if status != compat23_httplib.OK:
        raise CentralCLIAPIException('%s returned status code: %s' %
                                     (path, status))
    else:
        return request_id


def query_request(node, request_id):
    '''Query status of remote command invocation

    Return a tuple of request status and output from request.
    '''
    path = BASE_URL % 'query_request'
    status, json, text, resp = rest_utils.local_node_request(
        "POST",
        path=path,
        body={
            "node_uuid": node,
            "request_id": request_id,
        },
        proton_request=True)

    if status != compat23_httplib.OK:
        raise CentralCLIAPIException('%s returned status code: %s' %
                                     (path, status))
    else:
        return json['request_status'], json['output']


def _invoke_and_wait(node, cmd, modifier, context, username, rbac_permissions,
                     debug_mode):
    '''Invoke a cmd and block until it completes'''

    start_time = None
    # All cmds with 'get' should display datetime stamp in it's output
    if cmd.startswith("get"):
        start_time = time.time()

    request_id = invoke_command(node, cmd, modifier, context,
                                username, rbac_permissions, debug_mode)

    status, output = query_request(node, request_id)

    while (status in ('STATUS_RUNNING',
                      'STATUS_SCHEDULED',
                      'STATUS_CANCELING')):
        time.sleep(DELAY_BETWEEN_QUERY_REQUESTS)
        status, output = query_request(node, request_id)

    return status, node, output, modifier, start_time


def worker_init():
    # Ignore SIGINT in worker process
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def async_invoke_and_wait(nodes, cmd, modifier, context, username,
                          rbac_permissions, debug_mode, callback):
    '''Invoke command on nodes asynchrousnously and run callback on finish'''
    # Importing multiprocessing lazily, as ESX doesn't have this module as part
    # of its Python installation. This is never a problem as only managers need
    # multiprocessing module.
    from multiprocessing import Pool
    pool = Pool(ASYNC_POOL_SIZE, worker_init)

    results = [pool.apply_async(_invoke_and_wait,
                                args=(node, cmd, modifier, context,
                                      username, rbac_permissions, debug_mode),
                                callback=callback) for node in nodes]

    try:
        # If timeout is not specified, interrupt doesn't get delivered. This is
        # due to a Python bug http://bugs.python.org/issue8296. By specifying a
        # high timeout, we ensure that interrupt is received.
        [res.get(999999) for res in results]
    finally: #  Ensure that all pool workers exit
        # TODO Also request cancellation on remote node using cancel_request, if
        # the API call(s) didn't finish.
        pool.terminate()
        pool.join()


def _check_argument(node, action, context, arg_id, user_input, value, username,
                    debug_mode):
    import pwd
    status, json, text, resp = rest_utils.local_node_request(
        "POST",
        path=BASE_URL % 'check_argument',
        body={
            "node_uuid": node,
            "action": action,
            "context": context,
            "arg_id": arg_id,
            "user_input": user_input,
            "value": value,
            "username": str(pwd.getpwnam(username).pw_uid),
            "debug_mode": debug_mode,
        },
        proton_request=True)

    result = json['result']
    arguments = json['arguments']

    return result, arguments


def get_arg_values(node, context, arg_id, user_input, value, username, debug_mode):
    '''Get valid values for a enum or query arg'''
    result, arguments = _check_argument(node, 'get_arg_values', context,
                                        arg_id, user_input, value,
                                        username, debug_mode)
    if not result:
        raise CentralCLIAPIException('get_arg_values failed')
    else:
        return arguments


def verify_arg(node, context, arg_id, user_input, value, username, debug_mode):
    '''Verify a verify_arg'''
    result, arguments = _check_argument(node, 'verify_arg', context,
                                        arg_id, user_input, value,
                                        username, debug_mode)
    return result


def cancel_request(node, request_id):
    '''Cancel a remote command'''
    path = BASE_URL % 'cancel_request'
    status, json, text, resp = rest_utils.local_node_request(
        "POST",
        path=path,
        body={
            "node_uuid": node,
            "request_id": request_id,
        },
        proton_request=True)

    if status != compat23_httplib.OK:
        raise CentralCLIAPIException('%s returned status code: %s' %
                                     (path, status))
