# ***************************************************************************
# Copyright 2014-2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import featureswitch
import nsx_logging
import os
import subprocess
import sys

from cli.utils.constants import NODE_TYPE_CLOUD_SERVICE_MANAGER
from cli.utils.constants import NODE_TYPE_NSXAAS_DASHBOARD
from cli.utils.constants import NODE_TYPE_CONTAINER_KUBE_PROXY
from cli.utils.constants import NODE_TYPE_CONTAINER_NCP
from cli.utils.constants import NODE_TYPE_CONTAINER_NODE_AGENT
from cli.utils.constants import NODE_TYPE_CONTROLLER
from cli.utils.constants import NODE_TYPE_EDGE
from cli.utils.constants import NODE_TYPE_AUTONOMOUS_EDGE
from cli.utils.constants import NODE_TYPE_HOST_KVM
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_HOST_PUBLIC_CLOUD
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_HOST_WIN
from cli.utils.constants import NODE_TYPE_KEYMANAGER
from cli.utils.constants import NODE_TYPE_MANAGER
from cli.utils.constants import NODE_TYPE_POLICY
from cli.utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from cli.utils.constants import NODE_TYPE_PACE
from cli.utils.constants import NODE_TYPE_GLOBAL_MANAGER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_OEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.constants import NODE_TYPE_UNKNOWN
from cli.utils.location_utils import FilePaths
from cli.utils.error_code import ErrorCode as ec


lg = nsx_logging.getLogger("cli.utils.node_utils")


_RHEL_FILE= "/etc/redhat-release"
_SLES_FILE="/etc/SuSE-release"
_OEL_FILE="/etc/oracle-release"
_LSB_RELEASE_FILE= "/usr/bin/lsb_release"
# For ESXi and KVM, the node type is determined by querying "release" files
# or invoking "uname -a". However, we don't want to do this when get_node_type
# is called during a unit test, so the following flag allows unit tests to
# turn off query the system as needed.
_QUERY_SYSTEM_FOR_NODE_TYPE = True

EXPORT_TYPE_RESTRICTED = "restricted"
EXPORT_TYPE_UNRESTRICTED = "unrestricted"


def is_windows():
    """Returns if current platform is windows
    """
    return sys.platform == "win32"


def is_hypervisor():
    """Returns if node type is hypervisor
    """
    node_type = get_node_type()
    return node_type in (NODE_TYPE_HOST_SLESKVM, NODE_TYPE_HOST_CENTOSKVM, NODE_TYPE_HOST_ESX,
                         NODE_TYPE_HOST_RHELKVM, NODE_TYPE_HOST_UBUNTUKVM,
                         NODE_TYPE_HOST_WIN)


def is_baremetal():
    """Returns if node type is bare metal
    """
    node_type = get_node_type()
    return (node_type == NODE_TYPE_BAREMETAL_UBUNTU_SERVER or
            node_type == NODE_TYPE_BAREMETAL_RHEL_SERVER or
            node_type == NODE_TYPE_BAREMETAL_CENTOS_SERVER or
            node_type == NODE_TYPE_BAREMETAL_SLES_SERVER or
            node_type == NODE_TYPE_BAREMETAL_OEL_SERVER or
            node_type == NODE_TYPE_BAREMETAL_WINDOWS_SERVER or
            node_type == NODE_TYPE_BAREMETAL_RHEL_CONTAINER or
            node_type == NODE_TYPE_BAREMETAL_CENTOS_CONTAINER)


def is_appliance():
    """Returns if node type is an NSX appliance
    """
    node_type = get_node_type()
    return (node_type == NODE_TYPE_CONTROLLER or
            node_type == NODE_TYPE_EDGE or
            node_type == NODE_TYPE_AUTONOMOUS_EDGE or
            node_type == NODE_TYPE_KEYMANAGER or
            node_type == NODE_TYPE_PUBLIC_CLOUD_GATEWAY or
            node_type == NODE_TYPE_PACE or
            node_type == NODE_TYPE_GLOBAL_MANAGER or
            NODE_TYPE_MANAGER in node_type or
            NODE_TYPE_CLOUD_SERVICE_MANAGER in node_type or
            NODE_TYPE_NSXAAS_DASHBOARD in node_type or
            NODE_TYPE_POLICY in node_type)


def is_container():
    """Returns if node type is an NSX container
    """
    node_type = get_node_type()
    return (node_type == NODE_TYPE_CONTAINER_KUBE_PROXY or
            node_type == NODE_TYPE_CONTAINER_NCP or
            node_type == NODE_TYPE_CONTAINER_NODE_AGENT or
            NODE_TYPE_CONTAINER_KUBE_PROXY in node_type or
            NODE_TYPE_CONTAINER_NCP in node_type or
            NODE_TYPE_CONTAINER_NODE_AGENT in node_type)


def is_pace():
    """Returns if node is NSX PACE appliance
    """
    node_type = get_node_type()
    return (node_type == NODE_TYPE_PACE or
            NODE_TYPE_PACE in node_type)


def is_manager():
    """Returns True if node is a Manager type node (includes GMs)
    """
    node_type = get_node_type()
    return (node_type == NODE_TYPE_MANAGER or
            node_type == NODE_TYPE_GLOBAL_MANAGER or
            NODE_TYPE_MANAGER in node_type or
            NODE_TYPE_GLOBAL_MANAGER in node_type)


def _get_node_type_file():
    return FilePaths.NSX_ISSUE_FILE


# TODO: use GET /node when available
_NODE_TYPE = NODE_TYPE_UNKNOWN
def get_node_type():
    """Retrieve node type.

    Return:
        One of: NODE_TYPE_CLOUD_SERVICE_MANAGER,
                NODE_TYPE_CONTAINER_KUBE_PROXY, NODE_TYPE_CONTAINER_NCP,
                NODE_TYPE_CONTAINER_NODE_AGENT, NODE_TYPE_CONTROLLER,
                NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, NODE_TYPE_HOST_ESX,
                NODE_TYPE_HOST_PUBLIC_CLOUD, NODE_TYPE_HOST_RHELKVM,
                NODE_TYPE_HOST_UBUNTUKVM, NODE_TYPE_HOST_WIN, NODE_TYPE_PACE,
                NODE_TYPE_GLOBAL_MANAGER,
                NODE_TYPE_BAREMETAL_UBUNTU_SERVER,
                NODE_TYPE_BAREMETAL_RHEL_SERVER,
                NODE_TYPE_BAREMETAL_CENTOS_SERVER,
                NODE_TYPE_BAREMETAL_SLES_SERVER,
                NODE_TYPE_BAREMETAL_OEL_SERVER,
                NODE_TYPE_BAREMETAL_WINDOWS_SERVER,
                NODE_TYPE_BAREMETAL_RHEL_CONTAINER,
                NODE_TYPE_BAREMETAL_CENTOS_CONTAINER,
                NODE_TYPE_KEYMANAGER, NODE_TYPE_MANAGER, NODE_TYPE_POLICY,
                NODE_TYPE_NSXAAS_DASHBOARD, NODE_TYPE_PUBLIC_CLOUD_GATEWAY,
                NODE_TYPE_HOST_RHELKVM, NODE_TYPE_UNKNOWN
        Or, if an appliance has a combination of manager, csm, policy
        or dashboard roles, the return value can be a space delimited
        string with corresponding node IDs, for example:
        "manager csm policy dashboard"
    """
    global _NODE_TYPE
    if _NODE_TYPE != NODE_TYPE_UNKNOWN:
        return _NODE_TYPE
    # If node type not already already determined, inspect system to
    # to determine node type
    # Use /etc/nsx_issue to determine node type, note that on hosts the
    # node-type key does not differentiate between the different host
    # types so we have to use other means to determine the type of host
    nsx_issue_node_type = None
    try:
        with open(_get_node_type_file()) as fo:
            lines = fo.readlines()
        for line in lines:
            parts = line.split(":", 1)
            if parts[0].strip() == "node-type" and len(parts) > 1:
                nsx_issue_node_type = parts[1].strip()
                break
    except:
        lg.exception("Error processing nsx_issue file",
                     error_code=ec.FILE_IO_ERROR)
    if nsx_issue_node_type == NODE_TYPE_EDGE:
        _NODE_TYPE = NODE_TYPE_EDGE
    elif nsx_issue_node_type == NODE_TYPE_AUTONOMOUS_EDGE:
        _NODE_TYPE = NODE_TYPE_AUTONOMOUS_EDGE
    elif nsx_issue_node_type == NODE_TYPE_PUBLIC_CLOUD_GATEWAY:
        _NODE_TYPE = NODE_TYPE_PUBLIC_CLOUD_GATEWAY
    elif nsx_issue_node_type == NODE_TYPE_HOST_PUBLIC_CLOUD:
        _NODE_TYPE = NODE_TYPE_HOST_PUBLIC_CLOUD
    elif nsx_issue_node_type == NODE_TYPE_KEYMANAGER:
        _NODE_TYPE = NODE_TYPE_KEYMANAGER
    elif nsx_issue_node_type == NODE_TYPE_HOST_WIN:
        _NODE_TYPE = NODE_TYPE_HOST_WIN
    elif nsx_issue_node_type == NODE_TYPE_PACE:
        _NODE_TYPE = NODE_TYPE_PACE
    elif nsx_issue_node_type == NODE_TYPE_GLOBAL_MANAGER:
        _NODE_TYPE = NODE_TYPE_GLOBAL_MANAGER
    elif nsx_issue_node_type == NODE_TYPE_BAREMETAL_SERVER and is_windows():
        _NODE_TYPE = NODE_TYPE_BAREMETAL_WINDOWS_SERVER
    if nsx_issue_node_type is not None:
        node_roles = []
        if NODE_TYPE_MANAGER in nsx_issue_node_type:
            node_roles.append(NODE_TYPE_MANAGER)
        if NODE_TYPE_CLOUD_SERVICE_MANAGER in nsx_issue_node_type:
            node_roles.append(NODE_TYPE_CLOUD_SERVICE_MANAGER)
        if NODE_TYPE_NSXAAS_DASHBOARD in nsx_issue_node_type:
            node_roles.append(NODE_TYPE_NSXAAS_DASHBOARD)
        if NODE_TYPE_POLICY in nsx_issue_node_type:
            node_roles.append(NODE_TYPE_POLICY)
        if NODE_TYPE_CONTAINER_KUBE_PROXY in nsx_issue_node_type:
            node_roles.append(NODE_TYPE_CONTAINER_KUBE_PROXY)
        if NODE_TYPE_CONTAINER_NCP in nsx_issue_node_type:
            node_roles.append(NODE_TYPE_CONTAINER_NCP)
        if NODE_TYPE_CONTAINER_NODE_AGENT in nsx_issue_node_type:
            node_roles.append(NODE_TYPE_CONTAINER_NODE_AGENT)
        if NODE_TYPE_CONTROLLER in nsx_issue_node_type:
            node_roles.append(NODE_TYPE_CONTROLLER)
        if node_roles:
            if _NODE_TYPE != NODE_TYPE_UNKNOWN:
                node_roles.append(_NODE_TYPE)
            _NODE_TYPE = " ".join(node_roles)
    # Check for UbuntuKVM/RhelKVM/ESX last; otherwise, an NSX appliance
    # (which runs on Ubuntu, RHEL, or ESX) may be confused as host
    # UbuntuKVM, RhelKVM, or ESX.
    # CentOS and Rhel are binary compatible, align CentOS with Rhel Type
    if _NODE_TYPE == NODE_TYPE_UNKNOWN and _QUERY_SYSTEM_FOR_NODE_TYPE:
        if os.path.isfile(_RHEL_FILE)  and not os.path.isfile(_OEL_FILE):
            with open(_RHEL_FILE) as fd:
                version = fd.read()
                if "Red Hat Enterprise Linux" in version:
                    if nsx_issue_node_type == NODE_TYPE_HOST_KVM:
                        _NODE_TYPE = NODE_TYPE_HOST_RHELKVM
                    elif nsx_issue_node_type == NODE_TYPE_BAREMETAL_SERVER:
                        _NODE_TYPE = NODE_TYPE_BAREMETAL_RHEL_SERVER
                    elif nsx_issue_node_type == NODE_TYPE_BAREMETAL_CONTAINER:
                        _NODE_TYPE = NODE_TYPE_BAREMETAL_RHEL_CONTAINER
                elif "CentOS Linux" in version:
                    if nsx_issue_node_type == NODE_TYPE_HOST_KVM:
                        _NODE_TYPE = NODE_TYPE_HOST_CENTOSKVM
                    elif nsx_issue_node_type == NODE_TYPE_BAREMETAL_SERVER:
                        _NODE_TYPE = NODE_TYPE_BAREMETAL_CENTOS_SERVER
                    elif nsx_issue_node_type == NODE_TYPE_BAREMETAL_CONTAINER:
                        _NODE_TYPE = NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
        elif os.path.isfile(_SLES_FILE):
            with open(_SLES_FILE) as fd:
                version = fd.read()
                if "SUSE Linux" in version:
                    if nsx_issue_node_type == NODE_TYPE_HOST_KVM:
                        _NODE_TYPE = NODE_TYPE_HOST_SLESKVM
                    elif nsx_issue_node_type == NODE_TYPE_BAREMETAL_SERVER:
                        _NODE_TYPE = NODE_TYPE_BAREMETAL_SLES_SERVER
        elif os.path.isfile(_OEL_FILE):
            with open(_OEL_FILE) as fd:
                version = fd.read()
                if "Oracle Linux" in version:
                    if nsx_issue_node_type == NODE_TYPE_BAREMETAL_SERVER:
                        _NODE_TYPE = NODE_TYPE_BAREMETAL_OEL_SERVER
        elif os.path.isfile(_LSB_RELEASE_FILE) and os.access(
            _LSB_RELEASE_FILE, os.X_OK):
            command = ['lsb_release','-si']
            try:
                output = subprocess.check_output(command)
                output = output.decode("utf8")
                if "Ubuntu" in output:
                    if nsx_issue_node_type == NODE_TYPE_HOST_KVM:
                        _NODE_TYPE = NODE_TYPE_HOST_UBUNTUKVM
                    elif nsx_issue_node_type == NODE_TYPE_BAREMETAL_SERVER:
                        _NODE_TYPE = NODE_TYPE_BAREMETAL_UBUNTU_SERVER
            except subprocess.CalledProcessError: # non-zero return code
                pass
        else:
            command = ['/bin/uname','-a']
            try:
                output = subprocess.check_output(command)
                output = output.decode("utf8")
                if "VMkernel" in output:
                    _NODE_TYPE = NODE_TYPE_HOST_ESX
            except subprocess.CalledProcessError: # non-zero return code
                pass
    return _NODE_TYPE


def get_node_type_string():
    """Retrieve node type string.

    Return:
        One of: "Cloud Service Manager", "Controller", "Edge", "ESX",
                "Key Manager", "Manager", "NCP", "Policy",
                "Public Cloud Gateway", "Public Cloud Host",
                "RHEL KVM", "Ubuntu KVM", "Unknown", "Windows", "Autonomous Edge"
        Or, if an appliance has a combination of manager, csm or policy
        roles, the return value can be a comma delimited string
        with corresponding node strings, for example:
        "Manager, Cloud Service Manager, Policy"
    """
    result = "Unknown"
    node_type = get_node_type()
    if node_type == NODE_TYPE_KEYMANAGER:
        result = "Key Manager"
    elif node_type == NODE_TYPE_EDGE:
        result = "Edge"
    elif node_type == NODE_TYPE_AUTONOMOUS_EDGE:
        result = "Autonomous Edge"
    elif node_type == NODE_TYPE_PUBLIC_CLOUD_GATEWAY:
        result = "Public Cloud Gateway"
    elif node_type == NODE_TYPE_HOST_CENTOSKVM:
        result = "CentOS KVM"
    elif node_type == NODE_TYPE_HOST_SLESKVM:
        result = "SLES KVM"
    elif node_type == NODE_TYPE_HOST_ESX:
        result = "ESX"
    elif node_type == NODE_TYPE_HOST_PUBLIC_CLOUD:
        result = "Public Cloud Host"
    elif node_type == NODE_TYPE_HOST_RHELKVM:
        result = "RHEL KVM"
    elif node_type == NODE_TYPE_HOST_UBUNTUKVM:
        result = "Ubuntu KVM"
    elif node_type == NODE_TYPE_HOST_WIN:
        result = "Windows"
    elif node_type == NODE_TYPE_PACE:
        result = "Intelligence"
    elif node_type == NODE_TYPE_GLOBAL_MANAGER:
        result = "Global Manager"
    elif node_type == NODE_TYPE_BAREMETAL_UBUNTU_SERVER:
        result = "Ubuntu BareMetal Server"
    elif node_type == NODE_TYPE_BAREMETAL_RHEL_SERVER:
        result = "RHEL BareMetal Server"
    elif node_type == NODE_TYPE_BAREMETAL_CENTOS_SERVER:
        result = "CentOS BareMetal Server"
    elif node_type == NODE_TYPE_BAREMETAL_SLES_SERVER:
        result = "SUSE BareMetal Server"
    elif node_type == NODE_TYPE_BAREMETAL_OEL_SERVER:
        result = "Oracle BareMetal Server"
    elif node_type == NODE_TYPE_BAREMETAL_WINDOWS_SERVER:
        result = "Windows BareMetal Server"
    elif node_type == NODE_TYPE_BAREMETAL_RHEL_CONTAINER:
        result = "RHEL BareMetal Container"
    elif node_type == NODE_TYPE_BAREMETAL_CENTOS_CONTAINER:
        result = "Ubuntu BareMetal Container"
    node_roles = []
    if NODE_TYPE_MANAGER in node_type:
        node_roles.append("Manager")
    if NODE_TYPE_CLOUD_SERVICE_MANAGER in node_type:
        node_roles.append("Cloud Service Manager")
    if NODE_TYPE_NSXAAS_DASHBOARD in node_type:
        node_roles.append("NSXaas Dashboard")
    if NODE_TYPE_POLICY in node_type:
        node_roles.append("Policy")
    if NODE_TYPE_CONTAINER_KUBE_PROXY in node_type:
        node_roles.append("Kube Proxy")
    if NODE_TYPE_CONTAINER_NCP in node_type:
        node_roles.append("NCP")
    if NODE_TYPE_CONTAINER_NODE_AGENT in node_type:
        node_roles.append("Node Agent")
    if NODE_TYPE_CONTROLLER in node_type:
        node_roles.append("Controller")
    if node_roles:
        if result != "Unknown":
            node_roles.append(result)
        result = ", ".join(node_roles)
    return result


def get_node_version():
    """Returns the version present in the nsx_issue file
    """
    res = None
    try:
        with open(FilePaths.NSX_ISSUE_FILE, "r") as fd:
            lines = fd.readlines()
        for l in lines:
            if l.startswith("version"):
                res = l.split(':')[-1].strip()
    except IOError as e:
        lg.exception("Error reading nsx_issue file %s: %s"
                     % (FilePaths.NSX_ISSUE_FILE, e),
                     error_code=ec.IO_ERROR)
    finally:
        return res


def get_public_cloud_config():
    """Returns the information present in the public-cloud-config file

    Return:
        A dict containing keys for interface, mode, gw
    """
    res = {}
    try:
        with open(FilePaths.NSX_PUBLIC_CLOUD_CONFIG_FILE, "r") as fd:
            lines = fd.readlines()
        for ln in lines:
            ln = ln.strip()
            if ln.startswith('#'):
                # Line is a comment
                continue
            tmp = ln.split('=', 1)
            if tmp[0] == "interface":
                tmp_list = tmp[1].split(':', 2)
                res['interface'] = tmp_list[0]
                res['mode'] = tmp_list[1]
            if tmp[0] == "gw":
                res['gw'] = tmp[1]
    except IOError as e:
        lg.exception("Error reading public cloud config file %s: %s",
                     FilePaths.NSX_PUBLIC_CLOUD_CONFIG_FILE, e,
                     error_code=ec.IO_ERROR)
    finally:
        return res


def should_enable_central_cli():
    """Central CLI should only be enabled on nodes of having manager role"""
    node_type = get_node_type()
    return (NODE_TYPE_MANAGER in node_type or NODE_TYPE_GLOBAL_MANAGER in node_type)


def get_export_type():
    """Get export type.

    Returns:
        "restricted" or "unrestricted". Note, the UnrestrictedSKU feature
        switch is currently used to determin export type per BZ2106089.
        This method can also return None if an error occurs while accessing
        the feature switch.
    """
    try:
        return (EXPORT_TYPE_UNRESTRICTED
                if featureswitch.is_unrestricted_s_k_u_enabled() else
                EXPORT_TYPE_RESTRICTED)
    except Exception:
        lg.exception(
            "Error reading UnrestrictedSKU feature switch",
            error_code=ec.INTERNAL_ERROR)
    return None


def is_reboot_required():
    """Returns true if Esx hypervisor reboot is required
    """
    node_type = get_node_type()
    if node_type == NODE_TYPE_HOST_ESX:
        command = "vim-cmd hostsvc/hostsummary /| grep -i rebootRequired"
        try:
            output = subprocess.check_output(command, shell=True)
            output = output.decode("utf8")
            if "true" in output.lower():
                return True
        except subprocess.CalledProcessError: # non-zero return code
            pass
    else:
        lg.info("Host reboot criteria not applicable for host: %s", node_type)
    return False
