"""
A pure Python implementation of readline module.

This module has zero dependency on anything external, e.g. underlying readline
or libedit libraries. Primary motivation for writing this is lack of support
for callback from libedit based readline implementations and overall buggy
nature of libedit on ESX platform. Mac OS X also uses libedit as underlying
readline library. So hopefully this module will be helpful for that platform as
well.

What works?
Basic Emacs binding based cursor movements, history support etc.

What doesn't work?
- No support for Vi binding based movements.
- No support for history search

Readline is quite exhaustive library, and GNU readline implemention works quite
well. So this module is not trying to replace that. This is just an attempt to
fix problems we faced, while trying to work with libedit based readline.
"""
import functools
import nsx_logging
import re
from signal import signal, SIGPIPE, SIG_DFL
import string
import sys
import termios
import tty


lg = nsx_logging.getLogger("cli.utils.nsx_readline")

class UnsupportedKeyException(Exception):
    pass

class KeyBeyondBoundsException(Exception):
    pass

class NoHistoryException(Exception):
    pass

class Key(object):
    LF = '\x0d'
    CR = '\x0a'
    ENTER = '\x0d'
    BACKSPACE = '\x7f'
    SUPR = ''
    SPACE = '\x20'
    ESC = '\x1b'

    # CTRL
    CTRL_A = '\x01'
    CTRL_B = '\x02'
    CTRL_C = '\x03'
    CTRL_D = '\x04'
    CTRL_E = '\x05'
    CTRL_F = '\x06'
    CTRL_H = '\x08'
    CTRL_K = '\x0b'
    CTRL_L = '\x0c'
    CTRL_N = '\x0e'
    CTRL_P = '\x10'
    CTRL_Z = '\x1a'

    # ALT
    ALT_A = '\x1b\x61'

    # CTRL + ALT
    CTRL_ALT_A = '\x1b\x01'

    # cursors
    UP = '\x1b\x5b\x41'
    DOWN = '\x1b\x5b\x42'
    LEFT = '\x1b\x5b\x44'
    RIGHT = '\x1b\x5b\x43'

    CTRL_ALT_SUPR = '\x1b\x5b\x33\x5e'

    # other
    F1 = '\x1b\x4f\x50'
    F2 = '\x1b\x4f\x51'
    F3 = '\x1b\x4f\x52'
    F4 = '\x1b\x4f\x53'
    F5 = '\x1b\x4f\x31\x35\x7e'
    F6 = '\x1b\x4f\x31\x37\x7e'
    F7 = '\x1b\x4f\x31\x38\x7e'
    F8 = '\x1b\x4f\x31\x39\x7e'
    F9 = '\x1b\x4f\x32\x30\x7e'
    F10 = '\x1b\x4f\x32\x31\x7e'
    F11 = '\x1b\x4f\x32\x33\x7e'
    F12 = '\x1b\x4f\x32\x34\x7e'

    PAGE_UP = '\x1b\x5b\x35\x7e'
    PAGE_DOWN = '\x1b\x5b\x36\x7e'
    HOME = '\x1b\x5b\x48'
    END = '\x1b\x5b\x46'
    META_B = '\x1bb'
    META_F = '\x1bf'

    INSERT = '\x1b\x5b\x32\x7e'
    SUPR = '\x1b\x5b\x33\x7e'


class History(object):

    """CLI history manager for generic object"""

    def __init__(self):
        self.items = []
        self.cur = None # It means that we are not rendering history
        self.tmp = None # To store currently entered line by user

    def __str__(self):
        return "items: %s;cur: %s;tmp: %s" % (self.items, self.cur, self.tmp)

    def next_item(self):
        if self.cur is None: # Editing user entered command
            raise NoHistoryException

        # If we are at the most recent recorded history, return self.tmp
        if self.cur == len(self.items) - 1:
            self.cur = None
            return self.tmp
        else:
            self.cur += 1
            return self.items[self.cur]

    def prev_item(self, tmp_item=None):
        if (self.cur == 0) or (not self.items): # Oldest history, or no history
            raise NoHistoryException
        elif self.cur is None:
            self.cur = len(self.items) - 1
            self.tmp = tmp_item
        else:
            self.cur -= 1
        return self.items[self.cur]

    def add_item(self, item):
        if not item[1]: # Don't add if nothing entered by user
            return
        elif self.items and self.items[-1] == item: # Avoid duplicates
            self.cur = None
            return
        else:
            self.items.append(item)
            self.cur = None

class NSXReadline(object):

    def __init__(self):

        self.completer = lambda: None
        self.completion_display_matches_hook = self._completion_display_fn
        self._completion_char = None
        self.tab_chars = ('\t', '?')
        self.completion_chars = ('\n', '\r') + self.tab_chars
        self.history = History()
        self.cli_state = CLIState()
        # exclude '\r', '\n', '\x0c', '\t', '\x0b' from string.printable
        self.printable = "%s%s%s " % (string.digits,
                                      string.ascii_letters,
                                      string.punctuation)
        self.old_settings = None
        self.client_terminated = False

    def _completion_display_fn(self, substitution,
                               matches, longest_match_length):
        """Default completion display function"""
        str_output = ' '.join(matches)
        sys.stdout.write('\n')
        self.reset()
        sys.stdout.write("%s\n" % str_output)
        self.reset()

    def _get_matching_prefix(self, partial_input, matches):
        """When multiple matches found, autocomplete with common prefix.
        partial_input: input entered so far
        matches: list of matching subcommands
        """
        result = partial_input
        # size of last word, which is being autocompleted
        try:
            size = len(partial_input.split()[-1])
        except IndexError:
            size = 0

        s1 = min(matches, key=len)
        for i in range(size, len(s1)):
            for m in matches:
                if s1[i] != m[i]:
                    return result
            result = "%s%s" % (result, s1[i])

        return result

    def readkey(self):
        getchar = functools.partial(sys.stdin.read, 1)
        c1 = getchar()
        if not c1: # client terminated connection
            self.client_terminated = True
            return '\n'
        if ord(c1) != 0x1b:
            return c1
        c2 = getchar()
        if ord(c2) != 0x5b:
            return c1 + c2
        c3 = getchar()
        if ord(c3) != 0x33:
            return c1 + c2 + c3
        c4 = getchar()
        return c1 + c2 + c3 + c4

    def bell(self):
        sys.stdout.write('\a')
        sys.stdout.flush()

    def clear(self):
        """Clear all content on the screen"""
        sys.stdout.write('\033[H\033[J')
        sys.stdout.flush()

    def reset(self):
        """Bring the cursor to start of the prompt and delete all content"""
        sys.stdout.write('\r')
        sys.stdout.write('\033[K')
        sys.stdout.flush()

    def refresh(self, rewrite_prompt=False):
        """Rewrite the current line with expected characters."""
        total_backspaces = '\b' * (
            self.cli_state.prev_location - self.cli_state.start)
        # Move to the start of prompt
        sys.stdout.write(total_backspaces)
        # Delete everything until end of line
        sys.stdout.write('\033[K')
        if rewrite_prompt:
            sys.stdout.write('%s' % self.cli_state.prompt)
        # Render the expected input
        sys.stdout.write(self.get_input())
        total_backspaces = '\b' * (
            self.cli_state.length - self.cli_state.location)
        # Move the cursor to cli_state.location
        sys.stdout.write(total_backspaces)
        sys.stdout.flush()

    def get_line_buffer(self):
        return self.cli_state.user_input

    def get_input(self):
        # Return everything excluding prompt
        return self.cli_state.user_input

    def get_begidx(self):
        iter = re.finditer('\s', self.get_input())
        result = 0
        try:
            result = [m.start(0) for m in iter][-1] + 1
        except:
            pass
        return result

    def get_endidx(self):
        return (len(self.get_input()))

    def set_raw_mode(self):
        fd = sys.stdin.fileno()
        self.old_settings = termios.tcgetattr(fd)
        tty.setraw(sys.stdin.fileno())

    def unset_raw_mode(self):
        if self.old_settings:
            fd = sys.stdin.fileno()
            termios.tcsetattr(fd, termios.TCSADRAIN, self.old_settings)
            self.old_settings = None


    def raw_input(self, prompt='', initial_input=''):
        """
        Take the input and modify current_pos and line_buffer based on that
        """
        if self.client_terminated:
            raise EOFError
        self.cli_state = CLIState(prompt=prompt, user_input=initial_input)
        sys.stdout.write(self.cli_state.text)
        sys.stdout.flush()
        try:
            self.set_raw_mode()
        except:
            # Setting raw mode may fail for subprocess based interactions
            pass
        try:
            c = self.readkey()
            while c not in self.completion_chars:
                rewrite_prompt = False
                try:
                    self._completion_char = None
                    # Handle history within this class
                    if c in (Key.UP, Key.DOWN, Key.CTRL_P, Key.CTRL_N):
                        if c in (Key.UP, Key.CTRL_P):
                            new_cli_state = self.history.prev_item(
                                                (self.cli_state.prompt,
                                                 self.cli_state.user_input))
                        else:
                            new_cli_state = self.history.next_item()

                        self.cli_state.user_input = new_cli_state[1]
                        self.cli_state.location = self.cli_state.length
                    elif c == Key.CTRL_L:
                        # Clear the screen, including prompt. Afterwards refresh
                        # the screen, and rewrite prompt as well.
                        self.clear()
                        rewrite_prompt = True
                    else:
                        self.cli_state.keypress(c)
                except (UnsupportedKeyException, KeyBeyondBoundsException,
                        NoHistoryException):
                    self.bell()
                else:
                    self.refresh(rewrite_prompt)
                c = self.readkey()
        finally:
            self.unset_raw_mode()

        if c in self.tab_chars:
            # If user has entered tab as prior completion command, make
            # completion character as '?', otherwise keep it a tab.
            if c == '?' or self._completion_char == '\t':
                self._completion_char = '?'
            else:
                self._completion_char = '\t'

            state = 0
            output = []
            while True:
                result = self.completer(self.get_input(), state)
                state += 1
                if not result:
                    break
                else:
                    output.append(result)
            if not output:  # Sequence does not match any completion
                self.reset()
                return self.raw_input(prompt=prompt,
                                      initial_input=self.get_input())
            elif len(output) > 1:  # Multiple matches, don't autocomplete
                longest_match_length = max([len(i) for i in output])
                self.completion_display_matches_hook(
                    '', output, longest_match_length)
                common_prefix = self._get_matching_prefix(self.get_input(),
                                                          output)
                self.reset()
                return self.raw_input(prompt=prompt,
                                      initial_input=common_prefix)
            else:  # Just one match, autocomplete current input
                it = re.finditer('\s', self.get_input())
                space_locations = [m.start(0) for m in it]
                if space_locations:
                    last_space = space_locations[-1]
                    new_input = "%s%s" % (
                        self.get_input()[:(last_space + 1)], output[0])
                else:
                    new_input = output[0]

                self.reset()
                return self.raw_input(prompt=prompt, initial_input=new_input)
        elif c in ('\n', '\r'):
            self._completion_char = None  # Reset the completion character
            sys.stdout.write('\n')
            result = self.get_input()
            self.history.add_item((self.cli_state.prompt,
                                   self.cli_state.user_input.strip()))
            return result

    def set_completer(self, function):
        self.completer = function

    def set_completion_display_matches_hook(self, function):
        self.completion_display_matches_hook = function

    def parse_and_bind(self, *args, **kwargs):
        # This module is hard-coded to respond to '\t' and '?'
        pass

    def set_completer_delims(self, *args, **kwargs):
        # This module is hard-coded to have ' ' as delims
        pass

    def get_completion_type(self):
        return ord(self._completion_char)


class CLIState(object):
    """Manage the state of CLI session.
    A key press can be sent to this object to get a new CLI state. Accordingly
    this new state can be rendered by the client.
    """
    def __init__(self, prompt='', user_input=''):
        self.prompt = prompt
        self.user_input = user_input
        self.location = self.length # Cursor location
        self.no_op_keys = (Key.ESC, Key.CTRL_Z, Key.ALT_A, Key.CTRL_ALT_A,
                           Key.CTRL_ALT_SUPR, Key.F1, Key.F2, Key.F3, Key.F4,
                           Key.F5, Key.F6, Key.F7, Key.F8, Key.F9, Key.F10,
                           Key.F11, Key.F12, Key.PAGE_UP, Key.PAGE_DOWN,
                           Key.INSERT, Key.SUPR
                           )
        self.supported_keys = [Key.LF, Key.CR, Key.ENTER, Key.BACKSPACE,
                               Key.SPACE, Key.CTRL_A, Key.CTRL_B, Key.CTRL_C,
                               Key.CTRL_D, Key.CTRL_E, Key.CTRL_F, Key.CTRL_H,
                               Key.CTRL_K, Key.CTRL_P, Key.CTRL_N, Key.UP,
                               Key.DOWN, Key.LEFT, Key.RIGHT, Key.HOME, Key.END,
                               Key.META_B, Key.META_F]
        self.supported_keys.extend(list(string.printable))

    def __str__(self):
        return "%s:%s:%s" % (self.prompt, self.user_input, self.location)

    @property
    def text(self):
        """Total line text, including prompt and user_input"""
        return "%s%s" % (self.prompt, self.user_input)

    @property
    def length(self):
        """Total length of current CLI, which includes prompt and text"""
        return len(self.text)

    @property
    def start(self):
        """First location after prompt, where cursor can appear"""
        return len(self.prompt)

    @property
    def prev_location(self):
        """Previous location of the cursor"""
        try:
            return self._prev_location
        except AttributeError:
            return self.location

    @prev_location.setter
    def prev_location(self):
        raise Exception("Can't set prev_location outside of object")

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        if value > self.length or value < self.start:
            raise KeyBeyondBoundsException("Can't set cursor at: %s" % value)
        else:
            try:
                self._prev_location = self.location
                self._location = value
            except AttributeError:
                self._location = self._prev_location = value

    def keypress(self, key):
        if key not in self.supported_keys:
            raise UnsupportedKeyException
        elif key in (Key.LEFT, Key.CTRL_B):
            self.location -= 1
        elif key in (Key.RIGHT, Key.CTRL_F):
            self.location += 1
        elif key in (Key.BACKSPACE, Key.CTRL_H):
            self.user_input = "%s%s" % (self.text[self.start:self.location-1],
                                        self.text[self.location:])
            self.location -= 1
        elif (key == Key.CTRL_A) or (key == Key.HOME) or (key == Key.META_B):
            self.location = self.start
        elif (key == Key.CTRL_E) or (key == Key.END) or (key == Key.META_F):
            self.location = self.length
        elif key == Key.CTRL_C:
                raise KeyboardInterrupt
        elif key == Key.CTRL_D:
            if not self.user_input:
                # If there is nothing on command, exit
                raise EOFError
            else:
                raise UnsupportedKeyException
        elif key == Key.CTRL_K:
            self.user_input = self.text[self.start:self.location]
            # self.prev_location only gets updated when self.location gets an
            # update. Due to uniqueness of CTRL_K command, there is no change
            # in self.location. But we still need to somehow update
            # self.prev_location, otherwise it'd hold incorrect old value, prior
            # to entering CTRL_K command.
            self.location = self.location
        else:
            self.user_input = "%s%s%s" % (self.text[self.start:self.location],
                                          key,
                                          self.text[self.location:])
            self.location += 1
