# Copyright (C) 2019-2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import os
import platform
import traceback
import time
import uuid

import yaml
import sys
import nsx_logging
import subprocess
from sha.constants import CONDITION_WAVEFRONT, REACHED, UNREACHED_MSG

SHA_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
LOG_LEVEL = nsx_logging.INFO
SYSTEM = platform.system()

PLATFORM_ESX = "VMkernel"
PLATFORM_LINUX = "Linux"
PLATFORM_WINDOWS = "Windows"

DIST_ESX = "nsx-esx"
DIST_KVM = "nsx-kvm"
DIST_BMS = "nsx-bms"
DIST_EDGE = "nsx-edge"
DIST_UA = "nsx-manager nsx-policy-manager nsx-controller"
DIST_GM = "global-manager"
DIST_PUBLIC_CLOUD_GATEWAY = "nsx-public-cloud-gateway"
DIST_AUTONOMOUS_EDGE = "nsx-autonomous-edge"

PLUGIN_TYPE_LISTEN = "listen"

# nsx logging has been configured in preconfig.py
_logger = nsx_logging.getLogger(__name__)
_logger.info('config initializing')


for path in sys.path:
    _logger.info('path in sys.path: %s' % path)


# get a specific logger
def get_logger(name):
    return nsx_logging.getLogger(name)


# set memory usage soft limit
# soft, hard = resource.getrlimit(resource.RLIMIT_AS)
# resource.setrlimit(resource.RLIMIT_AS, (90 * 1024 ** 2, hard))

# retrieving HOST UUID
if platform.system() == PLATFORM_WINDOWS:
    # resolving NSX installation dir
    import winreg

    nsx_reg_path = r'SOFTWARE\VMWARE, Inc.\NSX'
    hKey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, nsx_reg_path)
    nsx_path, regtype = winreg.QueryValueEx(hKey, 'InstallDir')
    if regtype != winreg.REG_SZ:
        raise TypeError(
            'NSX installation dir registry is not a string, current value: {}'.format(nsx_path))
    NSX_DIR = nsx_path
    ISSUE_PATH = os.path.join(NSX_DIR, 'nsx_issue')
    HOST_CONFIG_PATH = os.path.join(NSX_DIR, 'host-cfg.xml')
else:
    ISSUE_PATH = r'/etc/nsx_issue'
    HOST_CONFIG_PATH = r'/etc/vmware/nsx/host-cfg.xml'

# get the node type (nsx-esx, nsx-kvm, nsx-bms, nsx-edge)
with open(ISSUE_PATH) as f:
    for line in f.readlines():
        if 'version' in line:
            VERSION = line.split(':')[-1].strip()
            continue
        if 'node-type' in line:
            NODE_TYPE = line.split(':')[-1].strip()
        if VERSION and NODE_TYPE:
            break
    else:
        _logger.error('can not determine which platform / version nsx sha is running on, exit now')
        raise Exception('can not determine which platform / version nsx sha is running on, exit now')

_logger.info("Got version {}, node-type: {}".format(VERSION, NODE_TYPE))


# must contain all these 3 items: nsx-manager nsx-policy-manager nsx-controller
def _is_ua_type(type):
    ua_types = {'nsx-manager', 'nsx-policy-manager', 'nsx-controller'}
    ts = type.split(' ')
    if len(ts) != len(ua_types):
        return False
    for t in ts:
        if t not in ua_types:
            return False
        ua_types.remove(t)
    return True

def _is_gm_type(type):
    return type == DIST_GM

if _is_ua_type(NODE_TYPE) or _is_gm_type(NODE_TYPE):
    HOST_CONFIG_PATH = r'/etc/vmware/nsx-appl-proxy/appl-proxy-public-cfg.xml'
    if _is_ua_type(NODE_TYPE):
       NODE_TYPE = DIST_UA
       _logger.info("It's an UA node.")
    else:
       _logger.info("It's an Global Manager node.")
       NODE_TYPE = DIST_GM


def is_ua():
    return NODE_TYPE == DIST_UA

def is_gm():
    return NODE_TYPE == DIST_GM

def node_type_is(*dist_types):
    return NODE_TYPE in dist_types


# waiting to retrieve host id from host-cfg.xml
should_sleep = False
while True:
    HOST_ID = None
    # noinspection PyBroadException
    try:
        import xml.etree.ElementTree
        HOST_ID = xml.etree.ElementTree.parse(HOST_CONFIG_PATH).find('uuid').text
        uuid.UUID(str(HOST_ID))
        _logger.info('Host id is {}.'.format(HOST_ID))
        break
    except Exception as e:
        should_sleep = True
        _logger.warn("Could not read a valid uuid from {}, "
                     "current value:{}, error: {}, retry in 10s".format(HOST_CONFIG_PATH, HOST_ID, e))
        time.sleep(10)

_host_id = int(HOST_ID.replace('-', ''), 16)
HOST_ID_LOW = (_host_id >> 64) & 0xffffffffffffffff
HOST_ID_HIGH = _host_id & 0xffffffffffffffff

if should_sleep:
    _logger.info("It's first time start, should wait 2 minutes to avoid race in MPA client.")
    time.sleep(120)

# get hostname
try:
    HOST_NAME = subprocess.check_output(
        "hostname",
        stderr=subprocess.PIPE,
        shell=True
    ).decode('utf-8').splitlines()[0].strip()
except Exception as e:
    _logger.info('could not get hostname, error: {}'.format(e))
    HOST_NAME = HOST_ID

_logger.info('host name is {}'.format(HOST_NAME))

# get TIME_THRESHOLD for utils run_command
TIME_THRESHOLD = 180
TIME_THRESHOLD_BOTTOM = 60
if SYSTEM == PLATFORM_ESX:
    try:
        sha_config_path = os.path.join(SHA_ROOT_DIR, 'sha_config.yml')
        with open(sha_config_path, 'r') as f:
            sha_config = yaml.load(f.read(), Loader=yaml.Loader)
            time_threshold = int(sha_config.get('TIME_THRESHOLD', None))
            if time_threshold > TIME_THRESHOLD_BOTTOM:
                TIME_THRESHOLD = time_threshold
                _logger.info('Got TIME_THRESHOLD from config:{}'.format(time_threshold))
            else:
                _logger.info('Invalid TIME_THRESHOLD from config: {}, '
                             'it should be a positive integer > {}.'.format(time_threshold, TIME_THRESHOLD_BOTTOM))
    except Exception as e:
        trace = traceback.format_exc()
        _logger.error('Cannot get a valid TIME_THRESHOLD from sha_config.yml,'
                      ' use the default {} instead. exception:{}, trace:{}'.format(TIME_THRESHOLD, e, trace))
_logger.info('Use TIME_THRESHOLD:{}'.format(TIME_THRESHOLD))

ENABLED_METRICS = {metric[0]: metric[1] for metric in [
    ['nsx.hyperbus.hyperbus-connection-list', 'HyperbusConnectionList', [DIST_ESX, DIST_KVM, ]],
    ['nsx.vmk.vmk-list', 'VMKList', [DIST_ESX, DIST_KVM, ]],
    ['nsx.daemon.daemon-health', 'DaemonHealth', [DIST_ESX, DIST_KVM, DIST_BMS, ]],
    ['nsx.ua.top-process-list', 'TopProcessList', [DIST_UA]],
    ['nsx.ua.ua-node-latency-list', 'UaNodeLatencyList', [DIST_UA]],
    ['nsx.host.host-metrics', 'HostMetrics', [DIST_ESX, ]],
    ['nsx.vdr.vdr-stats', 'VdrStats', [DIST_ESX, ]],
    ['nsx.edge.edge-fw-stats', 'EdgeFWStats', [DIST_EDGE, ]],
    ['nsx.edge.edge-fw-conn', 'EdgeFWConn', [DIST_EDGE, ]],
    ['nsx.edge.edge-ipsec-vpn-stats', 'EdgeIpsecVpnStats', [DIST_EDGE, ]],
    ['nsx.edge.edge-status', 'EdgeStatus', [DIST_EDGE, ]],
    ['nsx.portmirror.mirror-stack-status', 'MirrorStackStatus', [DIST_ESX, ]],
    ['nsx.auditlog.auditlog-health-for-local-error-status', 'AuditLogHealthForLocalErrorStatus',
     [DIST_ESX, DIST_KVM, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['nsx.auditlog.auditlog-health-for-remote-error-status', 'AuditLogHealthForRemoteErrorStatus',
     [DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['nsx.communication.manager-fqdn-lookup-failure-status', 'ManagerFqdnLookupFailureStatus',
     [DIST_BMS, DIST_ESX, DIST_KVM, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['nsx.communication.manager-fqdn-reverse-lookup-failure-status', 'ManagerFqdnReverseLookupFailureStatus',
     [DIST_UA, DIST_GM, ]],
    ['nsx.edge.edge-sr-state', 'EdgeSrState', [DIST_EDGE, ]],
    ['nsx.edge.mgmt-interface-status', 'EdgeMgmtInterface', [DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_AUTONOMOUS_EDGE, ]],
    ['nsx.edge.rcu-event-status', 'EdgeRcuBlock', [DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_AUTONOMOUS_EDGE, ]],
    ['nsx.lb-status', 'LbStatus', [DIST_EDGE, DIST_ESX,  ]],
] if NODE_TYPE in metric[2]}

ENABLED_MONITORS = {monitor[0]: monitor[1] for monitor in [
    ['UplinkStatusMonitor', 'nsx.uplink.uplink-status', [DIST_ESX, DIST_KVM, DIST_BMS]],
    ['ProcessStatusMonitor', ['nsx.process.nsx-cfgagent.liveness',
                              'nsx.process.nsx-opsagent.liveness',
                              'nsx.process.nsx-cfgagent.status.heapmemory.allocated',
                              'nsx.process.nsx-opsagent.status.heapmemory.allocated'],
     [DIST_ESX, DIST_EDGE, DIST_KVM, DIST_BMS, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
] if NODE_TYPE in monitor[2]}

ENABLED_ACTIONS = [action[0] for action in [
    ['Hbreport', [DIST_ESX, DIST_KVM, ]],
    ['Nareport', [DIST_ESX, DIST_KVM, ]],
    ['Dhreport', [DIST_ESX, DIST_KVM, DIST_BMS, ]],
    ['Alarm', [DIST_ESX, DIST_KVM, DIST_BMS, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['Exportdata', [DIST_EDGE, DIST_ESX]],
    ['Msreport', [DIST_ESX, ]],
    ['Tpreport', [DIST_UA]],
    ['UaNLReport', [DIST_UA]],
] if NODE_TYPE in action[1]]

PRE_INSTALL_SCRIPTS = {os.path.join(SHA_ROOT_DIR, 'plugins', plugin[0] + '.py'): plugin[0] for plugin in [
    ['hyperbus_monitor', [DIST_ESX, DIST_KVM, ]],
    ['node_agent_monitor', [DIST_ESX, DIST_KVM, ]],
    ['daemon_health_monitor', [DIST_ESX, DIST_KVM, DIST_BMS, ]],
    ['uplink_status_monitor', [DIST_ESX, DIST_KVM, DIST_BMS, ]],
    ['host_metrics_monitor', [DIST_ESX, ]],
    ['vdr_stats_monitor', [DIST_ESX, ]],
    ['edge_fw_stats_monitor', [DIST_EDGE, ]],
    ['edge_fw_conn_monitor', [DIST_EDGE, ]],
    ['edge_ipsec_vpn_stats_monitor', [DIST_EDGE, ]],
    ['edge_status_monitor', [DIST_EDGE, ]],
    ['mirrorstack_monitor', [DIST_ESX, ]],
    ['top_process_monitor', [DIST_UA]],
    ['audit_log_health_for_local_monitor', [DIST_ESX, DIST_KVM, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['audit_log_health_for_remote_monitor', [DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['manager_fqdn_lookup_failure_monitor', [DIST_BMS, DIST_ESX, DIST_KVM, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['manager_fqdn_reverse_lookup_failure_monitor', [DIST_UA, DIST_GM, ]],
    ['ua_node_latency_monitor', [DIST_UA]],
    ['edge_l3_monitor', [DIST_EDGE, ]],
    ['edge_appliance_monitor', [DIST_EDGE, ]],
    ['edge_mgmt_interface_monitor', [DIST_EDGE, DIST_AUTONOMOUS_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, ]],
    ['rcu_event_monitor', [DIST_EDGE, DIST_AUTONOMOUS_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, ]],
    ['process_status_monitor', [DIST_ESX, DIST_EDGE, DIST_KVM, DIST_BMS, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['service_opsagent_monitor', [DIST_ESX, ]],
    ['service_cfgagent_monitor', [DIST_ESX, ]],
    ['lb_status', [DIST_ESX, DIST_EDGE, ]],
] if NODE_TYPE in plugin[1]}

GLOBAL_MINIMUM_CHECKING_INTERVAL = 1
# max num for custom plugin
MAX_INSTALLED_CUSTOM_PLUGIN_NUM = 10

PRE_CONDITION = {
    CONDITION_WAVEFRONT: {REACHED: False, UNREACHED_MSG: 'Wavefront collector has not been set.'}
}

# Metrics, such as the ones use NsxxPRCClient, need use dedicated thread to handle
DEDICATED_THREAD_METRICS = ['nsx.daemon.daemon-health']

_logger.info('config initialized')
