# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import threading
import time
import traceback
import uuid
import xml.etree.ElementTree
import sha.constants
from sha import config
from sha.contrib.metric.utils import run_command


class ConfigMonitor(threading.Thread):
    """
    Monitor configuration changes
    """
    _logger = sha.config.get_logger(__name__)
    _CHECK_INTERVAL = 60  # seconds
    _host_id_lock = threading.Lock()

    def __init__(self):
        super().__init__(name='config-monitor')
        self._is_running = False

    def run(self):
        self._logger.info('ConfigMonitor starts to monitor...')
        self._is_running = True
        while self._is_running:
            try:
                self.check_host_id()
                self.check_host_name()
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning(
                    "Failed in monitoring config change. Exception: {}, trace:{}".format(e, trace))
            finally:
                self.sleep(self._CHECK_INTERVAL)
        self._logger.info('ConfigMonitor stopped monitoring.')

    def check_host_id(self):
        host_id = self.read_host_id()
        self._logger.debug('ConfigMonitor - got host uuid {}'.format(host_id))
        if host_id and host_id != config.HOST_ID:
            self.update_host_id(host_id)
            self._logger.info('ConfigMonitor - host uuid is updated to {}'.format(config.HOST_ID))

    def check_host_name(self):
        host_name = self.get_host_name()
        self._logger.debug('ConfigMonitor - got host name {}'.format(host_name))
        if host_name and host_name != config.HOST_NAME:
            config.HOST_NAME = host_name
            self._logger.info('ConfigMonitor - host name is updated to {}'.format(config.HOST_NAME))

    def read_host_id(self):
        host_id = None
        try:
            host_id = xml.etree.ElementTree.parse(config.HOST_CONFIG_PATH).find('uuid').text
            uuid.UUID(str(host_id))
        except Exception:
            trace = traceback.format_exc()
            self._logger.error('ConfigMonitor - could not read a valid uuid from:{}, current value:{}, trace:{}'
                               .format(config.HOST_CONFIG_PATH, host_id, trace))
            host_id = None
        return host_id

    @staticmethod
    def update_host_id(new_id):
        with ConfigMonitor._host_id_lock:
            config.HOST_ID = new_id
            _host_id = int(config.HOST_ID.replace('-', ''), 16)
            config.HOST_ID_LOW = (_host_id >> 64) & 0xffffffffffffffff
            config.HOST_ID_HIGH = _host_id & 0xffffffffffffffff

    @staticmethod
    def host_id_low_high():
        with ConfigMonitor._host_id_lock:
            return config.HOST_ID_LOW, config.HOST_ID_HIGH

    @staticmethod
    def host_id():
        with ConfigMonitor._host_id_lock:
            return config.HOST_ID

    def get_host_name(self):
        try:
            return run_command('hostname').splitlines()[0].strip()
        except Exception:
            trace = traceback.format_exc()
            self._logger.error('ConfigMonitor - failed in getting host name by cmd, use host id as name. '
                               'HOST_ID:{}, trace:{}'.format(config.HOST_ID, trace))
            return config.HOST_ID

    def stop(self):
        self._logger.warning("going to stop ConfigMonitor")
        self._is_running = False

    def sleep(self, seconds):
        sleep_seconds = seconds
        # check the running flag to exit in time
        while self._is_running and sleep_seconds > 0:
            time.sleep(1)
            sleep_seconds -= 1
