# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import sha.config

import gc
import os
import sys
import time
from threading import Lock, Thread
from sha.contrib.action.abc_base import Action
from sha.contrib.metric.utils import run_command
from sha.config_monitor import ConfigMonitor

from sha.contrib.action.wavefront_proxy_client import WavefrontProxyClient, metric_to_line_data

from xml.dom.minidom import Document

FIREWALL_RULESET_FILE = "/etc/vmware/firewall/nsx-sha-Ruleset.xml"
FIREWALL_REFRESH_CMD = ['localcli', 'network', 'firewall', 'refresh']
FIREWALL_ENABLE_CMD = ['localcli', 'network', 'firewall', 'ruleset', 'set', '-e', '1', '-r', 'nsx-sha']
FIREWALL_DISABLE_CMD= ['localcli', 'network', 'firewall', 'ruleset', 'set', '-e', '0', '-r', 'nsx-sha']

def putChildWithText(doc, parent, element, text):
    child = doc.createElement(element)
    child_text = doc.createTextNode(text)
    child.appendChild(child_text)
    parent.appendChild(child)

class WavefrontFirewall:
    """Firewall rules for wavefront.
       E.g:
        <ConfigRoot>
           <service id="0302">
              <id>nsx-sha</id>
              <rule id='0000'>
                 <direction>outbound</direction>
                 <protocol>tcp</protocol>
                 <porttype>dst</porttype>
                 <port>5002</port>
              </rule>
              <enabled>false</enabled>
              <required>false</required>
           </service>
        </ConfigRoot>
    """
    _logger = sha.config.get_logger(__name__)
    _wavefront_firewall_cleared = True
    _wavefront_firewall_file = FIREWALL_RULESET_FILE


    @classmethod
    def update(cls, portList):
        # empty portList, clear firewall rule
        if not portList:
            if not cls._wavefront_firewall_cleared:
                cls._wavefront_firewall_cleared = True
                try:
                    run_command(FIREWALL_DISABLE_CMD, shell=False)
                except Exception as e:
                    cls._logger.warning('Failed to disable firewall rule for wavefront:{}'.format(e))
            return

        #write firewall ruleset xml file
        doc = Document()
        root = doc.createElement('ConfigRoot')
        doc.appendChild(root)
        service = doc.createElement('service')
        root.appendChild(service)
        service.setAttribute('id','0301')
        putChildWithText(doc, service, 'id', 'nsx-sha')
        putChildWithText(doc, service, 'enabled', 'false')
        putChildWithText(doc, service, 'required', 'false')

        index = 0
        for port in portList:
            if port is None:
                continue
            rule = doc.createElement('rule')
            service.appendChild(rule)
            rule.setAttribute('id', '%04d' % index)
            putChildWithText(doc, rule, 'direction', 'outbound')
            putChildWithText(doc, rule, 'protocol', 'tcp')
            putChildWithText(doc, rule, 'porttype', 'dst')
            putChildWithText(doc, rule, 'port', '%d' % port)
            index += 1

        xml_str = doc.toprettyxml(indent='   ')
        cls._logger.debug('Firewall config: {}'.format(xml_str))

        with open(cls._wavefront_firewall_file, 'w') as f:
            f.write(xml_str)

        # update firewall rules
        cls._wavefront_firewall_cleared = False
        try:
            run_command(FIREWALL_REFRESH_CMD, shell=False)
            run_command(FIREWALL_ENABLE_CMD, shell=False)
        except Exception as e:
            cls._logger.warning('Failed to update firewall rule for wavefront:{}'.format(e))


class WavefrontSender:
    """The wavefront sender.
    """
    _logger = sha.config.get_logger(__name__)
    _wavefront_sender = None
    _wavefront_conf = None
    _wavefront_lock = Lock()


    @classmethod
    def _close_client(cls, stale_wavefront_sender):
        try:
            cls._logger.info('Close stale wavefront client: {}'.format(stale_wavefront_sender))
            stale_wavefront_sender.close()
            cls._logger.info('Stale wavefront client closed: {}'.format(stale_wavefront_sender))
            del stale_wavefront_sender
        except Exception as e:
            cls._logger.warning('Failed to close stale wavefront client:{}'.format(e))

    @classmethod
    def _clear(cls):
        with cls._wavefront_lock:
            cls._logger.info('Stale wavefront client: {}'.format(cls._wavefront_sender))
            stale_client = cls._wavefront_sender
            if cls._wavefront_sender is not None:
                cls._wavefront_sender = None

        if stale_client is not None:
            try:
                # Fire another thread to close stale wavefront client
                Thread(target=cls._close_client, args=[stale_client]).start()
            except Exception as e:
                cls._logger.warning('Failed to fire a thread to close stale wavefront client:{}'.format(e))

    @classmethod
    def update(cls, wavefront_conf):
        # update wavefront conf
        cls._logger.info('Update wavefront conf: {}'.format(wavefront_conf))
        cls._wavefront_conf = wavefront_conf

        # clear stale wavefront client
        cls._clear()

        # Create new wavefront client
        cls._logger.info('Create wavefront client to export data:')
        with cls._wavefront_lock:
            try:
                cls._wavefront_sender = WavefrontProxyClient(host=cls._wavefront_conf["host"],
                                                             metrics_port=cls._wavefront_conf["metrics_port"],
                                                             distribution_port=cls._wavefront_conf["distribution_port"],
                                                             tracing_port=cls._wavefront_conf["tracing_port"])
                cls._logger.info('Wavefront client created: {}'.format(cls._wavefront_sender))
            except Exception as e:
                cls._logger.warning('Failed to create wavefront client:{}'.format(e))
                cls._wavefront_sender = None

        # set firewall rule on ESX
        if sha.config.NODE_TYPE == sha.config.DIST_ESX:
            WavefrontFirewall.update([cls._wavefront_conf["metrics_port"],cls._wavefront_conf["distribution_port"],cls._wavefront_conf["tracing_port"]])


    @classmethod
    def clear(cls):
        # Clear wavefront conf
        cls._wavefront_conf = None

        # clear stale wavefront client
        cls._clear()

        # update firewall rule on ESX
        if sha.config.NODE_TYPE == sha.config.DIST_ESX:
            WavefrontFirewall.update([])

    @classmethod
    def isReady(cls):
        ret = False
        if cls._wavefront_lock.acquire(timeout = 1):
            if cls._wavefront_sender is not None:
                ret = True
            cls._wavefront_lock.release()
        else:
            cls._logger.warning('Failed to acquire lock to check wavefront client')
        return ret

    @classmethod
    def send_metric_now(cls, wavefront_data):
        if not cls._wavefront_lock.acquire(timeout = 1):
            cls._logger.warning('Failed to acquire lock before sending wavefront data')
            return
        try:
            if cls._wavefront_sender is None:
                cls._logger.warning('Wavefront collector has gone')
            else:
                cls._wavefront_sender.send_metric_now(wavefront_data)
                cls._logger.debug('Batch wavefront metric data sent: {}'.format(wavefront_data))
        except Exception as e:
            cls._logger.warning('Failed to send batch wavefront metric data: {}'.format(e))
            cls._logger.debug("Collecting...")
            try:
                n = gc.collect()
            except Exception as e:
                cls._logger.warning('Failed to collect garbage: {}'.format(e))
            cls._logger.debug("Number of unreachable objects collected by GC: {}".format(n))
            cls._logger.debug("Uncollectable garbage:{}".format(gc.garbage))
        finally:
            cls._wavefront_lock.release()


class Exportdata(Action):
    _logger = sha.config.get_logger(__name__)

    @classmethod
    def flat_nested_dict(cls, nested_dict):
        flat_dict = {}
        for (key, value) in nested_dict.items():
            if value is None:
                continue
            if type(value) is dict:
                new_dict = Exportdata.flat_nested_dict(value)
                for (new_key, new_value) in new_dict.items():
                   flat_dict['{0}.{1}'.format(key, new_key)] = new_value
            else:
                try:
                    flat_dict[key] = float(value)
                except Exception as e:
                    cls._logger.info('Failed to convert value:{}'.format(e))
                    pass

        return flat_dict

    @classmethod
    def populate_wavefront_data(cls, **kwargs):
        flat_dict = Exportdata.flat_nested_dict(kwargs['data'])
        wavefront_timestamp = int(kwargs.get('timestamp', time.time()))
        # Get source from kwargs, use NSX host_uuid if not set
        wavefront_source = kwargs.get('source', ConfigMonitor.host_id())
        batch_metric_data = []
        for (key, value) in flat_dict.items():
            one_metric_data = metric_to_line_data(
                name=key, value=value, timestamp=wavefront_timestamp,
                source=wavefront_source, tags={}, default_source="")
            batch_metric_data.append(one_metric_data)
        return batch_metric_data

    @classmethod
    def do(cls, hub, *args):
        def kernel(**kwargs):
            cls._logger.debug('signaling: {}'.format(kwargs))
            for export in kwargs['exports']:
                if export == 'wavefront':
                    wavefront_data = Exportdata.populate_wavefront_data(**kwargs)
                    cls._logger.debug('Wavefront metric data ready {}'.format(wavefront_data))
                    hub.send(wavefront_data)
                else:
                    cls._logger.debug('Unsupported export: {}'.format(export))

        return kernel
