# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import sha.config


import ctypes
import re
import socket
import sys
import time

def sanitize(string):
    """Sanitize a string, replace whitespace with "-".
    @param string: Input string
    @return: Sanitized string
    """
    whitespace_sanitized = re.sub(r'[\s]+', '-', string)
    if '"' in whitespace_sanitized:
        return '"' + re.sub(r'[\"]+', '\\\\\"', whitespace_sanitized) + '"'
    return '"' + whitespace_sanitized + '"'

def is_blank(string):
    """
    Check is a string is black or not, either none or only contains whitespace.
    @param string: String to be checked
    @return: Is blank or not
    """
    return string is None or len(string) == 0 or string.isspace()

def metric_to_line_data(name, value, timestamp, source, tags, default_source):
    """Metric Data to String.
    Wavefront Metrics Data format
    <metricName> <metricValue> [<timestamp>] source=<source> [pointTags]
    Example: "new-york.power.usage 42422 1533531013 source=localhost
              datacenter=dc1"
    @param name: Metric Name
    @type name: str
    @param value: Metric Value
    @type value: float
    @param timestamp: Timestamp
    @type timestamp: long
    @param source: Source
    @type source: str
    @param tags: Tags
    @type tags: dict
    @param default_source:
    @type default_source: str
    @return: String
    """
    if is_blank(name):
        raise ValueError('Metrics name cannot be blank')

    if is_blank(source):
        source = default_source

    str_builder = [sanitize(name), str(float(value))]
    if timestamp is not None:
        str_builder.append(str(int(timestamp)))
    str_builder.append('source=' + sanitize(source))
    if tags is not None:
        for key, val in tags.items():
            if is_blank(key):
                raise ValueError('Metric point tag key cannot be blank')
            if is_blank(val):
                raise ValueError('Metric point tag value cannot be blank')
            str_builder.append(sanitize(key) + '=' + sanitize(val))
    return ' '.join(str_builder) + '\n'


class PyObject(ctypes.Structure):
    _fields_ = [("refcnt", ctypes.c_long)]

class WavefrontProxyConn:
    def __init__(self, address, port,
                 timeout=None):
        """Construct ProxyConnectionHandler.
        @param address: Proxy Address
        @param port: Proxy Port
        @param timeout: Timeout to set on socket connecting to proxy
        """
        self._logger = sha.config.get_logger(__name__)
        self._address = address
        self._port = int(port)
        self.timeout = timeout
        self._reconnecting_socket = None


    def connect(self):
        """Initialize socket and connect to given address:port."""
        self._reconnecting_socket = socket.socket(socket.AF_INET,
                                                  socket.SOCK_STREAM)
        self._reconnecting_socket.settimeout(self.timeout)
        self._reconnecting_socket.connect((self._address, self._port))

    def close(self):
        """Close socket if it's open now."""
        if self._reconnecting_socket:
            self._logger.debug("{}, Refcount: {}".format(self._reconnecting_socket, sys.getrefcount(self._reconnecting_socket)))
            socket_addr = id(self._reconnecting_socket)
            self._reconnecting_socket.close()
            del self._reconnecting_socket
            self._logger.debug("Refcount after delete: {}".format(PyObject.from_address(socket_addr).refcnt))
            self._reconnecting_socket = None

    def send_data(self, line_data, reconnect=True):
        """Send data via proxy.

        @param line_data: Data to be sent
        @param reconnect: If it's the second time trying to send data
        """
        try:
            if not self._reconnecting_socket:
                self.connect()
            self._reconnecting_socket.sendall(line_data.encode('utf-8'))
        except socket.error as error:
            if reconnect:
                # Clear the old socket
                self.close()
                # Try to resend
                self.send_data(line_data, reconnect=False)
            else:
                # Second time trying failed
                raise error

class WavefrontProxyClient:
    def __init__(self, host, metrics_port, distribution_port=None, tracing_port=None,
                 timeout=None):
        """Construct wavefrontProxyClient.
        @param host: Proxy Host
        @param metrics_port: Proxy Metric Port
        @param distribution_port: Proxy Distribution Port, not implemented yet
        @param tracing_port: Proxy Tracing Port, not implemented yet
        @param timeout: Timeout to set on socket connecting to proxy
        """
        self._logger = sha.config.get_logger(__name__)
        self._host = host
        self._metrics_port = metrics_port
        self.timeout = timeout

        self._metrics_conn = (
            None if metrics_port is None
            else WavefrontProxyConn(host, metrics_port, timeout=timeout))

    def close(self):
        if self._metrics_conn:
            self._metrics_conn.close()

    def send_metric_now(self, metrics):
        if not self._metrics_conn:
            self._logger.warning("Can't send data to Wavefront. Please configure "
                                 "metrics port for Wavefront proxy.")
            return

        for metric in metrics:
            self._metrics_conn.send_data(metric)
