# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import json
import re
import sys
import traceback

from sha.contrib.metric.edge_base import EdgeBase
from sha.contrib.metric.utils import timeout_cache

class EdgeFWConn(EdgeBase):

    def get_fw_conn(self, interface):
        fw_conn = {}

        # Get firewall connection raw data.
        #
        # The cost of nsxcli is much higher than edge backend command
        #     """nsxcli -c 'get firewall {} connection raw | json'"""
        # nsxcli output:
        # [
        #   {
        #     "conn": "0x0000002e48000005 af 2  ethertype 0x0000 proto tcp   40.40.40.1:37849 (40.40.40.1:37849) -> 40.40.40.10:179 dir 2 0 0 0 0 state 4:4 f-1025 n-0 flg:200000040000"
        #   },
        #   {
        #     "conn": "0x0000002e48000007 af 2  ethertype 0x0000 proto tcp   40.40.40.10:40480 -> 40.40.40.1:179 (40.40.40.1:179) dir 1 0 0 0 0 state 4:4 f-1025 n-0 flg:200000040000"
        #   }
        #   {
        #     "conn": "0x0000002e7000002f af 2  ethertype 0x0000 proto icmp  2.1.4.10 -> 1.1.1.10 (1.1.1.10) dir 1 0 0 0 0 f-1025 n-0 flg:200000040000"
        #   }
        # ]
        #
        # Edge backend command output:
        # {"ifuuid":"eeb35dd2-8e1f-43b1-bb64-6bac81a35d82","vrf":1,"sync-enabled":"false","pbr-flag":"false","rule-flag":"false","ike-flag":"false","iftype":"DOWNLINK","config-loaded":"true","name":"plr1_downlink_1","parent-name":"DR-Tier0-LR-1","iface-of":"fca9fcbb-ed5f-4f44-bc4e-a4c91704ee1d","pvi":2,"connections":{"flg-hex":"sloppy|sync|uplink|vti syn|invalid|-|- -|procl7|ls|rulel7 l7|-|closing|expiring 0 lookup 0 ref hicore0 locore0 hicore1 locore1","connection-states":"
        # 0x0000002e7000033e af 2  ethertype 0x0000 proto tcp   40.40.40.10:58192 -> 40.40.40.1:179 (40.40.40.1:179) dir 1 28 0 1cb0a 0 state 4:4 f-1025 n-0 flg:200000040000
        # 0x0000002e7000033f af 2  ethertype 0x0000 proto icmp  2.1.4.10 -> 1.1.3.10 (1.1.3.10) dir 1 0 0 0 0 f-1025 n-0 flg:200000040000
        # 0x0000002e70000340 af 2  ethertype 0x0000 proto icmp  1.1.2.10 -> 2.1.4.10 (1.1.2.10) dir 2 0 0 0 0 f-1025 n-0 flg:200000040000
        # 0x0000002e70000341 af 2  ethertype 0x0000 proto icmp  1.1.1.10 -> 2.1.4.10 (1.1.1.10) dir 2 0 0 0 0 f-1025 n-0 flg:200000040000
        # "},"connection-count":"4","TCP Half Opened Active/Max":"0/1000000","UDP Active/Max":"0/100000","ICMP Active/Max":"1/10000","Other Active/Max":"0/10000"}

        cmd_get_conn = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'fw/show', interface, 'connections']
        output = self.run_command_with_time_check(cmd_get_conn)
        self._logger.debug("firewall conn via backend: {}".format(output))
        connections = re.search('"connection-states":"([^"]*)"', output).group(1).strip().splitlines()
        self._logger.debug("firewall conn: {}".format(connections))

        # Calculate flowsPerCore, estConn and tcpConn
        flowsPerCore = {}
        # set default conn count per-proto as 0
        estConn = {'tcp': 0, 'udp':0, 'others':0}
        fw_conn['tcp'] = 0

        for item in connections:
            match = re.search(r'proto (\S*) .*dir (\d) [ a-f\d]* ([a-z]\S*) (\S*).* flg:([a-f\d]*)', item)
            if match is None:
                self._logger.warning("Failed to parse conn{}".format(item['conn']))
                continue
            proto = match.group(1)

            # flowsPerCore
            dir = int(match.group(2))
            flag = match.group(5)
            txCore = 'c' + '%x' %int(flag[-2:], 16)
            rxCore = 'c' + '%x' %int(flag[-4:-2], 16)
            self._logger.debug("get rxCore:{}, txCore:{}".format(rxCore, txCore))
            if dir == 1:
                # IN
                if rxCore in flowsPerCore:
                    flows = flowsPerCore[rxCore]
                    flows['Rx'] = flows.get('Rx', 0) + 1
                else:
                    flowsPerCore[rxCore] = {'Rx': 1}
            elif dir == 2:
                # OUT
                if txCore in flowsPerCore:
                    flows = flowsPerCore[txCore]
                    flows['Tx'] = flows.get('Tx', 0) + 1
                else:
                    flowsPerCore[txCore] = {'Tx': 1}
            else:
               self._logger.warning("Unknown direction {} in flow: {}".format(dir, item))

            # tcp connetion count
            if proto == 'tcp':
                fw_conn['tcp'] = fw_conn.get('tcp', 0) + 1

            # estConnection count
            self._logger.debug("Match result for state: {}, {}".format(match.group(3), match.group(4)))
            if match.group(3) != 'state':
                continue
            state = match.group(4)
            if state == "4:4" or state == "EST:EST":
                if proto not in ['tcp', 'udp', 'icmp']:
                     proto = 'others'
                estConn[proto] = estConn.get(proto, 0) + 1

        if flowsPerCore:
            fw_conn['flowsPerCore'] = flowsPerCore
        if estConn:
            fw_conn['estConn'] = estConn


        self._logger.debug("get edge fw conn({}): {}".format(sys.getsizeof(fw_conn), fw_conn))
        return fw_conn

    @timeout_cache(10)
    def usage(self, *args, **kwargs):
        lr_list = None
        try:
            lr_list = self.get_fw_interface_list()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get edge fw interface for T0/T1 lrouter: {}, {}".format(e, trace))
        finally:
            if lr_list is None:
                return None

        lr_conn = {}
        for lr in lr_list:
            if 'interface' not in lr:
                continue

            try:
                fw_conn = self.get_fw_conn(lr['interface'])
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get edge fw conn on {}: {}, {}".format(lr['interface'], e, trace))
                fw_conn = None

            if fw_conn is not None:
                lr_conn[lr['type'] + '.' +lr['name']] = fw_conn

        return None if not lr_conn else (self.wavefront_source, lr_conn)
