# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential

import time
import traceback

from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.schema.open_vswitch import impl_idl

import sha.config
from sha.contrib.monitor.abc_base import Monitor
from sha.contrib.monitor.uplinkstatus.ovs.notification_handle import handle_notification

if sha.config.SYSTEM == sha.config.PLATFORM_WINDOWS:
    SOCK_FILE_PATH = sha.config.NSX_DIR + "\\openvswitch\\var\\run\\openvswitch\\db.sock"
else:
    SOCK_FILE_PATH = "/var/run/openvswitch/db.sock"
OVS_SCHEMA_NAME = "Open_vSwitch"
OVSDB_SERVER_ADDR = "unix:" + SOCK_FILE_PATH


class OvsUplinkStatusMonitor(Monitor):
    """
    OvsUplinkStatusMonitor monitors the uplink state change on dataplane which
    are running OVS.
    """

    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self._connection = None
        self._api = None
        self._status_cache = dict()
        self.is_running = False
        self._stop = False

    def start(self):
        while not self._stop:
            delta_wait = 10
            try:
                idl = connection.OvsdbIdl.from_server(OVSDB_SERVER_ADDR, OVS_SCHEMA_NAME)
                self._connection = connection.Connection(idl, 5)
                self._api = impl_idl.OvsdbIdl(self._connection)
                if not self._api:
                    self._logger.error("Init ovsdbapp failed - not get api,  will retry...")
                    time.sleep(delta_wait)
                    continue
                self._logger.info("Init ovsdbapp success")
                self._api.idl.notify = self.handle_notification
                self.is_running = True
                self._logger.info("OvsUplinkStatusMonitor started.")
                return
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.error("OvsUplinkStatusMonitor failed in start, will retry. Exception:{} trace:{}.".format(e, trace))
                time.sleep(delta_wait)

    def handle_notification(self, event, row, updates=None):
        try:
            status_changes = dict()
            handle_notification(self._api, status_changes, event, row, updates)
            for alarm, status in status_changes.items():
                if status != self._status_cache.get(alarm, None):
                    self._status_cache[alarm] = status
                    self.notify({alarm: status})
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.error("Handle notification failed from ovsdb-server, exception:{}, trace:{}".format(e, trace))

    def stop(self):
        self._logger.info("Stopping OvsUplinkStatusMonitor.")
        self._stop = True
        if self.is_running:
            try:
                self._connection.stop()
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.error("Error in stop ovs db connection, Exception:{}, trace:{}".format(e, trace))
        self._logger.info("OvsUplinkStatusMonitor is stopped.")
