# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import json
import os
import ipaddress
from vmware.nsx.nestdb import nestdb_schema_common_pb2 as schema_common_pb2
from sha.contrib.action.exportdata import WavefrontSender
import sha.config
import yaml
from sha.config import PRE_CONDITION
from sha.constants import CONDITION_WAVEFRONT, REACHED


class CollectorResolver:
    _logger = sha.config.get_logger(__name__)
    _collectors = dict()
    _mapper = {
           schema_common_pb2.NetopaCollectorMsg.CollectorType.Value('WAVE_FRONT'): 'wavefront',
    }

    @classmethod
    def get(cls, type):
        """
        get the collector of the specified type
        """
        return cls._collectors[type]

    @classmethod
    def _set(cls, collector_msg):
        """
        :param collector_msg: protobuf_msg
        """
        sha_collector_config = collector_msg.netopa_collector_config
        for item in sha_collector_config.collector:
            cls._logger.debug("item in collector msg: {}".format(item))
            if item.type is None:
                continue
            type = cls._mapper.get(item.type)
            if type == "wavefront":
                collector = cls._collectors.setdefault(type, Collector(type))
                if item.collector_ip is not None:
                    collector.set('ip', item.collector_ip)
                if item.collector_port is not None:
                    collector.set('port', item.collector_port)
                collector.update()
                cls._logger.debug("wavefront collector:{}".format(collector))
                # The wavefront collector is updated, skip the loop
                return True
        return False

    @classmethod
    def _clear(cls, old_collector_msg):
        """
        :param old_collector_msg: protobuf_msg
        """
        old_collector_config = old_collector_msg.netopa_collector_config
        for item in old_collector_config.collector:
            cls._logger.debug("item in old collector msg: {}".format(item))
            if item.type is None:
                continue
            type = cls._mapper.get(item.type)
            collector = cls._collectors.setdefault(type, Collector(type))
            collector.clear()

    @classmethod
    def set(cls, collector_msg, old_collector_msg):
        """
        :param collector_msg: protobuf_msg
        :param old_collector_msg: protobuf_msg
        """
        if cls._set(collector_msg) is not True:
            cls._clear(old_collector_msg)
        cls._logger.debug(cls._collectors)

    @classmethod
    def clear(cls, old_collector_msg):
        """
        :param old_collector_msg: protobuf_msg
        """
        cls._clear(old_collector_msg)


wavefront_default_collector = {
    'host' : None,
    'metrics_port' : 2878,
    'distribution_port' : None,
    'tracing_port': None}

class Collector:
    """The collector config.
    """
    def __init__(self, type):
        self._logger = sha.config.get_logger(__name__)
        self._type = type
        self._attrs = dict()
        self._changed = False

    def set(self, attr, val):
        if (attr not in self._attrs) or \
           (self._attrs[attr] != val):
            self._changed = True
            self._attrs[attr] = val

    def get(self):
        return self._attrs

    def update(self):
        if self._changed is not True:
            return
        if self._type == "wavefront":
            try:
                # Update wavefront conf
                self._logger.info('Update wavefront collector: {}'.format(self._attrs))
                wavefront_conf = wavefront_default_collector
                if 'ip' in self._attrs:
                    if self._attrs['ip'].ipv4 is not None:
                        wavefront_conf['host'] = str(ipaddress.IPv4Address(self._attrs['ip'].ipv4))
                    elif self._attrs['ip'].ipv6 is not None:
                        wavefront_conf['host'] = str(ipaddress.IPv6Address(self._attrs['ip'].ipv6))
                if 'port' in self._attrs:
                    wavefront_conf["metrics_port"] = int(self._attrs['port'])
                self._logger.info(wavefront_conf)
                # Update wavefront sender
                WavefrontSender.update(wavefront_conf)
                # Update pre_conditioin
                if not PRE_CONDITION[CONDITION_WAVEFRONT][REACHED]:
                    PRE_CONDITION[CONDITION_WAVEFRONT][REACHED] = True
                    self._logger.info("sha precondition changed: {}".format(PRE_CONDITION))
            except Exception as e :
                import traceback; traceback.print_exc(); print(e)
        self._changed = False

    def clear(self):
        self._attrs.clear()
        if self._type == "wavefront":
            WavefrontSender.clear()
            if PRE_CONDITION[CONDITION_WAVEFRONT][REACHED]:
                PRE_CONDITION[CONDITION_WAVEFRONT][REACHED] = False
                self._logger.info("sha precondition changed: {}".format(PRE_CONDITION))

