# ***************************************************************************
# Copyright 2020-2021 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import errno
import os
import pty
import re
import sys
import shutil
from subprocess import Popen
from subprocess import PIPE

_NSX_COMMON_PYTHON_PATH = "/opt/vmware/nsx-common/python"
if _NSX_COMMON_PYTHON_PATH not in sys.path:
    sys.path.append(_NSX_COMMON_PYTHON_PATH)
import nsx_logging


# ###########################################################################
# IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT
# ###########################################################################
#
# If you are updating this file, keep in mind that this file is used by NSX
# services running on hosts (e.g. ESX, RHEL, Ubuntu) registered with the MP.
# This means that:
#
# 1) NAPI may not be running; therefore, you cannot import NAPI modules.
#
# 2) Python 3 may be used; therefore, your code has to be both Python 2 & 3
#    compatible.
#
# ###########################################################################
try:
    import gevent.os
    do_os_read = gevent.os.nb_read
    do_os_write = gevent.os.nb_write
    do_fd_nonblocking = gevent.os.make_nonblocking
except ImportError:
    do_os_read = os.read
    do_os_write = os.write
    do_fd_nonblocking = None


lg = nsx_logging.getLogger("napi.root.node.file-store.utils")


# Supported fingerprint hashes/digests defined in openssh digest.h
_SUPPORTED_FINGERPRINT_HASHES = [
    "MD5", "RIPEMD160", "SHA1", "SHA256", "SHA384", "SHA512"]

_SSH_VERSION_RE = re.compile("^OpenSSH_([0-9\.]+).+$")


# This file cannot import NAPI modules per IMPORTANT notice above; therefore,
# we are setting _INTERNAL_ERROR here to match the value defined by NAPI in
# node/api/src/python/management_api/napi/error_code.py
_INTERNAL_ERROR = "NODE1"


# This file may be used on Hypervisors where simplejson hasn't been installed
try:
    import simplejson
except ImportError:
    import json as simplejson


# This file may be used on Hypervisors where Python 3 is installed instead
# of Python 2.
if sys.version_info < (3,):
    from cStringIO import StringIO
    import httplib
    _HTTPLIB_BAD_REQUEST = httplib.BAD_REQUEST
    _HTTPLIB_UNAUTHORIZED = httplib.UNAUTHORIZED
    _HTTPLIB_NOT_FOUND = httplib.NOT_FOUND
    _HTTPLIB_INTERNAL_SERVER_ERROR = httplib.INTERNAL_SERVER_ERROR
    _HTTPLIB_SERVICE_UNAVAILABLE = httplib.SERVICE_UNAVAILABLE
    _HTTPLIB_GATEWAY_TIMEOUT = httplib.GATEWAY_TIMEOUT
else:
    from io import StringIO
    import http.client
    _HTTPLIB_BAD_REQUEST = http.client.BAD_REQUEST
    _HTTPLIB_UNAUTHORIZED = http.client.UNAUTHORIZED
    _HTTPLIB_NOT_FOUND = http.client.NOT_FOUND
    _HTTPLIB_INTERNAL_SERVER_ERROR = http.client.INTERNAL_SERVER_ERROR
    _HTTPLIB_SERVICE_UNAVAILABLE = http.client.SERVICE_UNAVAILABLE
    _HTTPLIB_GATEWAY_TIMEOUT = http.client.GATEWAY_TIMEOUT


FILE_STORE_DIR = "/var/vmware/nsx/file-store"
# This is currently only used by NSX Inteligence node.
FILE_STORE_DATA_DIR = "/data/vmware/nsx/file-store"
SIG_NO_SPACE = "No space left on device"
SIG_NO_SPACE_SFTP = "Couldn't write to remote file"
DEFAULT_FILESTORE_FILES = ["backup_restore_helper.py",
                           "aggsvc_poll_intervals_change_helper.py",
                           "nsx_backup_cleaner.py", "get_backup_timestamps.sh"]

_SCP_CMD = "/usr/bin/scp"
_SFTP_CMD = "/usr/bin/sftp"
_SSH_CMD = "/usr/bin/ssh"

_CONNECT_TIMEOUT = 3  # seconds

_DEFAULT_PORT_PROTOCOL_MAP = {
    "http": 80,
    "https": 443,
    "scp": 22,
    "sftp": 22}

# match both hex string (md5, older versions of openssh) and base64
# (newer versions of openssh)
_FINGERPRINT_RE = re.compile("^.+ key fingerprint is ([0-9a-zA-Z:+/]+)\..*$")
_SFTP_RE = re.compile("^sftp", re.MULTILINE)

_ERR_MSG_BAD_FINGERPRINT = "Invalid fingerprint specified."
_ERR_MSG_CONNECTION_ERROR = "Error or timeout when connecting to remote server."
_ERR_MSG_FILE_NOT_FOUND = "Specified uri not found."
_ERR_MSG_GENERIC = "Error copying remote file."
_ERR_MSG_NO_DISKSPACE = "Insufficient diskspace to fulfill request."
_ERR_MSG_UNATHORIZED = "The remote server returned unauthorized error."
_ERR_MSG_NEGOTIATION_ERROR = "Error negotiating with remote host."

_PRIORITIZED_CIPHERS = (
    "chacha20-poly1305@openssh.com,aes256-gcm@openssh.com,"
    "aes128-gcm@openssh.com,aes256-ctr,aes192-ctr,aes128-ctr,"
    "aes256-cbc,aes192-cbc,aes128-cbc,3des-cbc")


def cleanup_filestore():
    files_found = []
    success = True
    for filename in os.listdir(FILE_STORE_DIR):
        if filename in DEFAULT_FILESTORE_FILES:
            continue
        path = os.path.join(FILE_STORE_DIR, filename)
        try:
            if os.path.isdir(path):
                shutil.rmtree(path)
            if os.path.isfile(path):
                os.unlink(path)
        except OSError as ex:
            if hasattr(ex, "errno") and ex.errno == errno.ENOENT:
                pass
            else:
                success = False
                lg.warn("Unable to delete %s: %s" % (path, ex))
    return success


def invoke_command(cmd_list, stdin=None, env=None):
    """Invoke specified cmd_list with optional stdin and env

    Returns:
        (returncode, stdout, stderr) tuple where stdout and stderr values
        have been normalized to strings (i.e. account for Python3
        subprocess.communicate returning byte strings).
    """
    if sys.version_info >= (3,) and stdin is not None:
        # In Python3, communicate() expects stdin as a byte string
        stdin = bytes(stdin, "utf8")
    po = Popen(
        cmd_list, stdin=(PIPE if stdin else None), stdout=PIPE, stderr=PIPE,
        env=env)
    out, err = po.communicate(input=stdin)
    if sys.version_info >= (3,):
        # In Python3, communicate() returns out and err as byte strings
        if out is not None:
            out = out.decode("utf8")
        if err is not None:
            err = err.decode("utf8")
    return po.returncode, out, err


def ssh_copy_command_list(protocol, port, remote_username, src, dst, host_key_algorithms=None,
                          preserve_file_properties=True):
    """Constucts scp/sftp command to invoke to copy to/from remote file.

    Args:
        protocol: scp or sftp.
        port: Port number to which to connect.
        remote_username: Username to authenticate with on remote server.
        src: Source file to copy from. This is either a string which implies
            a local file or a (ip_address/hostname, uri) tuple which implies
            a remote file.
        dst: Destination file to copy to. This is either a string which implies
            a local file or a (ip_address/hostname, uri) tuple which implies
            a remote file.
        host_key_algorithms: The host key algorithms the client wants to use.

    Returns:
        (copy_cmd_list, sftp_put_cmd) tuple to be passed to invoke_ssh_cmd.
        sftp_put_cmd is None if the protocol is not sftp or if a sftp GET
        operation is specified. (Syntax of sftp GETs are the same as scp.)
    """
    def _remote_file(username, server, uri=None):
        if uri:
            uri = uri if uri.startswith("/") else "/%s" % uri
            return "%s@%s:%s" % (username, server, uri)
        else:
            return "%s@%s" % (username, server)

    sftp_put_cmd = None
    if isinstance(src, tuple):
        src_file = _remote_file(remote_username, src[0], src[1])
        dst_file = dst
    else:
        if protocol == "scp":
            src_file = src
            dst_file = _remote_file(remote_username, dst[0], dst[1])
        else:
            src_file = None
            dst_file = _remote_file(remote_username, dst[0])
            dst_uri = dst[1] if dst[1].startswith("/") else "/%s" % dst[1]
            sftp_put_cmd = "PUT %s %s" % (src, dst_uri)
    cmd = _SCP_CMD if protocol == "scp" else _SFTP_CMD
    # Set PreferredAuthentications to include both password and
    # keyboard-interactive to account for non-password authentication
    # that the remote server may prefer (e.g. if it sets
    # "PasswordAuthentication no" in its /etc/ssh/sshd_config file).
    cmd_list = [
        cmd, "-P", str(port), "-o", "NumberOfPasswordPrompts=1",
        "-o", "PreferredAuthentications=password,keyboard-interactive",
        "-o", "ConnectTimeout=%s" % _CONNECT_TIMEOUT,
        "-c", _PRIORITIZED_CIPHERS]
    if preserve_file_properties:
       cmd_list.append("-p")
    if host_key_algorithms:
        cmd_list.append("-o")
        cmd_list.append("HostKeyAlgorithms=%s" % host_key_algorithms)
    if src_file:
        cmd_list.append(src_file)
    cmd_list.append(dst_file)
    return cmd_list, sftp_put_cmd


def ssh_mkdir_command_list(protocol, port, remote_username, dst):
    """Constucts sftp command to invoke to make remote directory

    Args:
        protocol: sftp.
        port: Port number to which to connect.
        remote_username: Username to authenticate with on remote server.
        dst:  Destination host/directory to create, of the form
        (ip_address/hostname, directory).

    Returns:
        (copy_cmd_list, sftp_put_cmd) tuple to be passed to invoke_ssh_cmd.
    """
    sftp_put_cmd = None
    if protocol == "sftp":
        server = "%s@%s" % (remote_username, dst[0])
        dst_uri = dst[1] if dst[1].startswith("/") else "/%s" % dst[1]
        sftp_put_cmd = "MKDIR %s" % (dst_uri)
    else:
        raise NotImplementedError("%s protocol not supported" % protocol)

    cmd = _SFTP_CMD
    # Set PreferredAuthentications to include both password and
    # keyboard-interactive to account for non-password authentication
    # that the remote server may prefer (e.g. if it sets
    # "PasswordAuthentication no" in its /etc/ssh/sshd_config file).
    cmd_list = [
        cmd, "-P", str(port), "-o", "NumberOfPasswordPrompts=1",
        "-o", "PreferredAuthentications=password,keyboard-interactive",
        "-o", "ConnectTimeout=%s" % _CONNECT_TIMEOUT]
    cmd_list.append(server)
    return cmd_list, sftp_put_cmd


def ssh_ls_command_list(protocol, port, remote_username, dst):
    """Constucts sftp command to invoke to list remote directory

    Args:
        protocol: sftp.
        port: Port number to which to connect.
        remote_username: Username to authenticate with on remote server.
        dst:  Destination host/directory to list, of the form
        (ip_address/hostname, directory).

    Returns:
        (copy_cmd_list, sftp_put_cmd) tuple to be passed to invoke_ssh_cmd.
    """
    sftp_put_cmd = None
    if protocol == "sftp":
        server = "%s@%s" % (remote_username, dst[0])
        dst_uri = dst[1] if dst[1].startswith("/") else "/%s" % dst[1]
        sftp_put_cmd = "LS -l %s" % (dst_uri)
    else:
        raise NotImplementedError("%s protocol not supported" % protocol)

    cmd = _SFTP_CMD
    # Set PreferredAuthentications to include both password and
    # keyboard-interactive to account for non-password authentication
    # that the remote server may prefer (e.g. if it sets
    # "PasswordAuthentication no" in its /etc/ssh/sshd_config file).
    cmd_list = [
        cmd, "-P", str(port), "-o", "NumberOfPasswordPrompts=1",
        "-o", "PreferredAuthentications=password,keyboard-interactive",
        "-o", "ConnectTimeout=%s" % _CONNECT_TIMEOUT]
    cmd_list.append(server)
    return cmd_list, sftp_put_cmd



def invoke_ssh_cmd(req, cmd_list, sftp_put_cmd, expected_fingerprint,
                   remote_password, server, port, uri,
                   error_response_method=None):
    """Execute specified ssh/scp/sftp command, handle fingerprint
       verification as well as password entry.

    Args:
        req: The HTTP request, only used if error_response_method is not None.
        cmd_list: List containing command and args to invoke as returned by
            ssh_copy_command_list.
        sftp_put_cmd: sftp command to invoke returned by ssh_copy_command_list.
        expected_fingerprint: Expected fingerprint which should already be
            in username's known_hosts file or returned by the remote server
            if not already in known_hosts file.
        remote_password: Password to use when authenticating to remote
            server.
        server: Remote server.
        port: Remote server port.
        uri: URI of remote server file.
        error_response_method: Reference to RestRoutineProvider.error_response.
            This is an optional argument to allow callers the option of
            getting a text error description instead of a webob response
            object.
    Returns:
        (True, None) if ssh command successful.
        (False, webob Error Response) if ssh command failed and
            error_response_method is not None.
        (False, error text) if ssh command failed and error_response_method
            is None.
    """
    try:
        cmd_list = list(cmd_list)  # Make a copy given we're about to update it
        # The filestore APIs require that the SSH fingerprint be specified
        # on each invocation. Therefore, we specify /dev/null as the
        # known_hosts file to force a fingerprint check each time.
        pos = cmd_list.index("-o")
        cmd_list.insert(pos, "UserKnownHostsFile=/dev/null")
        cmd_list.insert(pos, "-o")
        lg.debug("invoke cmd_list %s", cmd_list)
        lg.debug("invoke ssh %s", sftp_put_cmd)
        status, out = _ssh_cmd(
            cmd_list, sftp_put_cmd, expected_fingerprint, remote_password)
        # If our first attempt resulted in a fingerprint error and if
        # expected_fingerprint is prefixed with a hash algorithm, invoke
        # the ssh command again with the -o FingerprintHash option using
        # the specified hash algorithm
        if (status is None and out is None and
                _ssh_supports_fingerprint_hash_option()):
            hash_algorithm, fp = _split_ssh_fingerprint(expected_fingerprint)
            pos = cmd_list.index("-o")
            cmd_list.insert(pos, "FingerprintHash=%s" % hash_algorithm)
            cmd_list.insert(pos, "-o")
            expected_fingerprint = ":".join([hash_algorithm, fp])
            status, out = _ssh_cmd(
                cmd_list, sftp_put_cmd, expected_fingerprint, remote_password)
        return _process_ssh_result(
            req, status, out, sftp_put_cmd is not None, server, port, uri,
            error_response_method)
    except Exception:
        lg.exception("Error copying remote filestore file",
                     error_code=_INTERNAL_ERROR)
    if error_response_method is None:
        return False, _ERR_MSG_GENERIC
    return False, error_response_method(
        req, _HTTPLIB_INTERNAL_SERVER_ERROR,
        "NODE_ERROR_COPYING_REMOTE_FILE")


def _process_ssh_result(req, status, out, is_sftp_op, server, port, uri,
                        error_response_method):
    """Process _ssh_cmd result.

    Args:
        req: The HTTP request, only used if error_response_method is not None.
        status: status value returned by _ssh_cmd method.
        out: out value returned by _ssh_cmd method.
        is_sftp_op: True, if _ssh_cmd invoke with SFTP operation
        server: Remote server.
        port: Remote server port.
        uri: URI of remote server file.
        error_response_method: Reference to RestRoutineProvider.error_response.
            This is an optional argument to allow callers the option of
            getting a text error description instead of a webob response
            object.
    Returns:
        (True, None) if ssh command successful.
        (False, webob Error Response) if ssh command failed and
            error_response_method is not None.
        (False, error text) if ssh command failed and error_response_method
            is None.
    """
    # If not SFTP operation, status == 0 implies a successful operation
    if not is_sftp_op and status == 0:
        return True, None
    # If no status or out value, then a connection was never establish
    elif status is None or out is None:
        if error_response_method is None:
            return False, _ERR_MSG_BAD_FINGERPRINT
        return False, error_response_method(
            req, _HTTPLIB_BAD_REQUEST,
            "NODE_ERROR_FILESTORE_REMOTE_SERVER_FINGERPRINT_MISMATCH")
    # Parse _ssh_cmd out value to determine whether an error occurred
    elif "Permission denied" in out:
        if error_response_method is None:
            return False, _ERR_MSG_UNATHORIZED
        return False, error_response_method(
            req, _HTTPLIB_UNAUTHORIZED,
            "NODE_ERROR_FILESTORE_REMOTE_SERVER_UNAUTHORIZED")
    # String "Bad message" will be in output, when sftp path invalid
    elif ("No such file" in out) or ("Bad message" in out):
        if error_response_method is None:
            return False, _ERR_MSG_FILE_NOT_FOUND
        return False, error_response_method(
            req, _HTTPLIB_NOT_FOUND,
            "NODE_ERROR_FILESTORE_REMOTE_FILE_NOT_FOUND", uri)
    # Check of ssh negotiation error
    elif (out.startswith("Unable to negotiate with ") or
            "forbidden in FIPS mode" in out):
        if error_response_method is None:
            return False, _ERR_MSG_NEGOTIATION_ERROR
        return False, error_response_method(
            req, _HTTPLIB_BAD_REQUEST,
            "NODE_ERROR_FILESTORE_REMOTE_HOST_NEGOTIATION_FAILED", out)
    elif out == "" and (status == 256 or (is_sftp_op and status != 0)):
        # Connection timeouts return status 256 with no output.
        if error_response_method is None:
            return False, _ERR_MSG_CONNECTION_ERROR
        return False, error_response_method(
            req, _HTTPLIB_GATEWAY_TIMEOUT,
            "NODE_ERROR_FILESTORE_REMOTE_SERVER_CONNECTION_ERROR",
            server if port == 22 else "%s:%s" % (server, port))
    elif SIG_NO_SPACE in out or (is_sftp_op and SIG_NO_SPACE_SFTP in out):
        # Out of diskspace error returns text error message
        if error_response_method is None:
            return False, _ERR_MSG_NO_DISKSPACE
        return False, error_response_method(
            req, _HTTPLIB_SERVICE_UNAVAILABLE,
            "COMMON_NO_DISKSPACE_LEFT_ERROR")
    elif is_sftp_op and status == 0:
        # A SFTP operation didn't return any of the known error output,
        # assume success.
        return True, None
    # None of the known errors detected, return generic error message
    if error_response_method is None:
        return False, _ERR_MSG_GENERIC
    return False, error_response_method(
        req, _HTTPLIB_BAD_REQUEST,
        "NODE_ERROR_REMOTE_SERVER_ERROR",
        "status: %s   out: '%s'" % (status, out))


def _ssh_cmd(cmd_list, sftp_put_cmd, expected_fingerprint, remote_password):
    """Execute specified ssh/scp/sftp command, handle fingerprint
       verification as well as password entry.

    Args:
        cmd_list: List containing command and args to invoke.
        sftp_put_cmd: sftp command to invoke.
        expected_fingerprint: Expected fingerprint which should already be
            in username's known_hosts file or returned by the remote server
            if not already in known_hosts file.
        remote_password: Password to use when authenticating to remote
            server.
    Returns:
        (exit_status, output) tuple. Note, if ssh returns a fingerprint
        that doesn't match expected_fingerprint, (None, None) is returned.
    """
    def os_read(fd, n, strip=False):
        # os.read, but return string instead of bytes in Python 3
        output = do_os_read(fd, n)
        if output is None:
            return None
        elif sys.version_info >= (3,):
            output = output.decode("utf8", "backslashreplace")
            if "\\u" in output:
                lg.info("Non-UTF8 encoding detected from ssh response: '%s'", output)
        if strip:
            return output.strip()
        return output

    def os_write(fd, buf):
        # os.write accounting for Python 3 which writes bytes intead of strings
        if sys.version_info >= (3,):
            do_os_write(fd, bytes(buf, "utf8"))
        else:
            do_os_write(fd, buf.encode("utf8"))

    pid, fd = pty.fork()
    if not pid:
        # Replace child process with specified ssh/scp command
        cmd_list = [item.encode("utf8") for item in cmd_list]
        os.execv(cmd_list[0], cmd_list)
    # Assume ssh always prompts for password and/or to confirm new host
    if do_fd_nonblocking is not None:
        do_fd_nonblocking(fd)
    buffer_output = ""
    ssh_os_error = False
    ssh_os_error_msg = ""
    fingerprint_match = None
    expected_fingerprint = expected_fingerprint.replace(":", "")
    server_fingerprint = ""
    while True:
        try:
            output = os_read(fd, 1024, True)
            buffer_output += output
            if " password:" in buffer_output or "Password:" in buffer_output:
                os_write(fd, remote_password + "\n")
                break
            else:
                # If ssh output is declaring the remote server's fingerprint,
                # the current user's known_hosts file doesn't know about the
                # the remote server; therefore, compare expected fingerprint
                # with the remote server's fingerprint
                if fingerprint_match is None:
                    match = _FINGERPRINT_RE.match(buffer_output.replace("\n", ""))
                    if match:
                        server_fingerprint = match.group(1).replace(":", "")
                        if (server_fingerprint.lower() !=
                                expected_fingerprint.lower()):
                            fingerprint_match = False
                        else:
                            fingerprint_match = True
                if "(yes/no)?" in output:
                    # Write yes or no based on fingerprint match result
                    if not fingerprint_match:
                        if fingerprint_match is None:
                            lg.warning("Rejecting the remote server because "
                                "fingerprint is missing. SSH output received: "
                                "%s", str(buffer_output))
                        else:
                            lg.warning("Rejecting the remote server because "
                                "fingerprint is invalid, expected: %s, "
                                "actual: %s", expected_fingerprint.lower(),
                                server_fingerprint.lower())
                        os_write(fd, "no\n")
                        os.close(fd)
                        os.waitpid(pid, 0)
                        return None, None
                    os_write(fd, "yes\n")
                else:
                    if (output.startswith("Unable to negotiate with ") or
                          "forbidden in FIPS mode" in output):
                        return -1, output
                    elif (not output.startswith("Warning: Permanently added") and
                            not output.startswith("Could not create directory") and
                            not output.startswith("yes") and
                            not output.startswith("The authenticity of host") and
                            "key fingerprint" not in output and
                            bool(output.strip())):
                        lg.error("Unexpected ssh output: %s" % output,
                                 error_code=_INTERNAL_ERROR)
        except OSError as e:
            lg.exception("OSError processing ssh command %s", e,
                         error_code=_INTERNAL_ERROR)
            # Keep record of the OSError message
            ssh_os_error = True
            ssh_os_error_msg = str(e)
            break
        except Exception as e:
            lg.exception("Error processing ssh command %s", e,
                         error_code=_INTERNAL_ERROR)
    exit_status = -1
    result = ""
    buf = StringIO()
    sftp_command_submitted = False
    try:
        # In case of OSError, use the exception message as output
        if ssh_os_error:
            buf.write(ssh_os_error_msg)
        else:
            while True:
                try:
                    output = os_read(fd, 1024)
                    lg.debug("output %s", output)
                    if not output:
                        break
                    buf.write(output)
                    if sftp_put_cmd and _SFTP_RE.search(output):
                        if sftp_command_submitted:
                            os_write(fd, "exit\n")
                            lg.debug("sent exit")
                        else:
                            os_write(fd, "%s\n" % sftp_put_cmd)
                            lg.debug("sent command %s", sftp_put_cmd)
                            sftp_command_submitted = True
                except OSError:
                    break
        try:
            _, exit_status = os.waitpid(pid, 0)
            # The real exit_status can be stored in the high byte, extract it
            if exit_status > 0x00FF:
                exit_status = (exit_status & 0xFF00) >> 8
        except OSError as ex:
            lg.exception("OSError on waitpid after ssh command %s", ex,
                         error_code=_INTERNAL_ERROR)
            # SCP child process exits early
            if hasattr(ex, "errno") and ex.errno == errno.ECHILD:
                exit_status = 256
            else:
                raise
        result = buf.getvalue()
    finally:
        try:
            # Close fd if not already closed, we could call os.fstat to see
            # if it's already closed, but that throws an exception also.
            os.close(fd)
        except:
            pass
        buf.close()
    return (exit_status, result)


def get_remote_server_auth_credentials(req_obj):
    """Return credentials to use when authenticating with remote server.

    Args:
        req_obj: Python dictionary containing request body properties.
    Returns:
        (scheme, username, password) if specified in request, or
        (None, None, None) if not specified.
    """
    obj = req_obj.get("protocol", {}).get("authentication_scheme", {})
    scheme = obj.get("scheme_name")
    username = obj.get("username")
    password = obj.get("password")
    return scheme, username, password


def get_remote_server_properties(req_obj):
    """Return remote server's ip_addr/hostname, port, protocol if specified.

    Args:
        req_obj: Python dictionary contain request body properties.
    Returns:
        (server ip_addr/hostname, port, protocol) if specified in request,
        note any of the tuple items can be none if not specified, e.g. if
        request does not require a remote host or if port is not specified.
    """
    protocol = req_obj.get("protocol", {}).get("name")
    server = req_obj.get("server")
    port = req_obj.get("port")
    if port is None and protocol:
        port = _DEFAULT_PORT_PROTOCOL_MAP.get(protocol)
    return (server, port, protocol)


def _split_ssh_fingerprint(fingerprint):
    """Return (hash_algorithm, fingerprint) tuple of specified ssh fingerprint.

    Args:
        fingerprint: Fingerprint to split
    Returns:
        (hash_algorithm, fingerprint) tuple where hash_algorithm can be
        one of _SUPPORTED_FINGERPRINT_HASHES. If specified fingerprint
        is not prefixed with one of _SUPPORTED_FINGERPRINT_HASHES, a
        legacy MD5 fingerprint is assumed. An example of a legacy
        fingerprint is "aa:15:8a:03:16:bb:06:0c:0a:42:db:c6:8b:14:9a:67"
    """
    parts = fingerprint.split(":", 1)
    if len(parts) == 2 and parts[0].upper() in _SUPPORTED_FINGERPRINT_HASHES:
        return parts[0].upper(), parts[1]
    # Assume legacy MD5 fingerprint
    return ("MD5", fingerprint)


def _ssh_supports_fingerprint_hash_option():
    rc, out, err = invoke_command([_SSH_CMD, "-V"])
    if rc != 0:
        return False
    m = _SSH_VERSION_RE.match(err)
    if not m:
        # Check stdout output in case not found in stderr output
        m = _SSH_VERSION_RE.match(out)
    if m:
        ver_str = m.group(1)
        ver_tuple = tuple([int(item) for item in ver_str.split(".")])
        # FingerprintHash option support starts in 6.8
        # per http://www.openssh.com/txt/release-6.8
        if ver_tuple >= (6, 8):
            return True
    return False


if __name__ == "__main__":
    # Copy local file to a SCP or SFTP server.
    #
    # Details of the local file and SCP/SFTP server are specified via stdin
    # via the following python dictionary format serialized using json:
    # {
    #   "local_path": "/home/admin/somefile.txt",
    #   "server": "10.1.1.1",
    #   "port": 22,
    #   "uri": "/tmp/somefile.txt",
    #   "protocol": {
    #     "name": "scp",
    #     "ssh_fingerprint": "b5:08:df:c6:55:62:e4:6e:95:70:7c:25:ba:f2:46:f1",
    #     "authentication_scheme": {
    #       "scheme_name": "password",
    #       "username": "root",
    #       "password": "123456"
    #     }
    #   }
    # }
    nsx_logging.basicConfig(stdout=True)
    try:
        json_input = sys.stdin.read()
        req_obj = simplejson.loads(json_input)
        path = req_obj.get("local_path")
        server, port, protocol = get_remote_server_properties(req_obj)
        scheme, usr, pwd = get_remote_server_auth_credentials(req_obj)
        fingerprint = req_obj.get(
            "protocol", {}).get("ssh_fingerprint", "")
        cmd_list, sftp_put_cmd = ssh_copy_command_list(
            protocol, port, usr, path, (server, req_obj["uri"]))
        success, err_msg = invoke_ssh_cmd(
            None, cmd_list, sftp_put_cmd, fingerprint, pwd, server, port,
            req_obj["uri"], error_response_method=None)
        if not success:
            sys.stderr.write(err_msg)
    except Exception:
        success = False
        err_msg = "Error copying remote file"
        sys.stderr.write(err_msg)
        lg.exception("Error copying remote file, json_input: %s", json_input,
                     error_code=_INTERNAL_ERROR)
    sys.exit(0 if success else 1)
