# ***************************************************************************
# Copyright 2017-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import grp
import os
import pwd
import shutil
from subprocess import PIPE
from subprocess import Popen
import sys
import traceback

sys.path.append("/opt/vmware/nsx-common/python")

import nsx_logging


# ###########################################################################
# IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT
# ###########################################################################
#
# If you are updating this file, keep in mind that this file is used by NSX
# services running on hosts (e.g. ESX, RHEL, Ubuntu) registered with the MP.
# This means that Python 3 may be used; therefore, your code has to be both
# Python 2 & 3 compatible.
#
# ###########################################################################
def invoke_command(cmd_list, stdin=None, env=None):
    """Invoke specified cmd_list with optional stdin and env

    Returns:
        (returncode, stdout, stderr) tuple where stdout and stderr values
        have been normalized to strings (i.e. account for Python3
        subprocess.communicate returning byte strings).
    """
    if sys.version_info >= (3,) and stdin is not None:
        # In Python3, communicate() expects stdin as a byte string
        stdin = bytes(stdin, "utf8")
    po = Popen(
        cmd_list, stdin=(PIPE if stdin else None), stdout=PIPE, stderr=PIPE,
        env=env)
    out, err = po.communicate(input=stdin)
    if sys.version_info >= (3,):
        # In Python3, communicate() returns out and err as byte strings
        if out is not None:
            out = out.decode("utf8")
        if err is not None:
            err = err.decode("utf8")
    return po.returncode, out, err


_LOG_TIMEOUT = 300  # Abort log collection via API after this num seconds
_NAPI_DIR = "/opt/vmware/nsx-node-api/bin/python"
_NSX_PYTHON_COMMON = "/opt/vmware/nsx-common/python"
_PYTHONPATH = _NAPI_DIR + os.path.pathsep + _NSX_PYTHON_COMMON
_NAPI_AUDIT_LOG_SCRIPT = (
    "%s/management_api/napi/root/administration/audit_logs/__self__.py" %
    _NAPI_DIR)

_PLATFORM_USR = "nsxplatform"
_PLATFORM_GRP = "nsxplatform"

_NODETYPE_MANAGER = "nsx-manager"
_NODETYPE_CONTROLLER = "nsx-controller"
_NODETYPE_EDGE = "nsx-edge"
_NODETYPE_PUBLIC_CLOUD_GATEWAY = "nsx-public-cloud-gateway"
_NODETYPE_UNKNOWN = "Unknown"

_OK = 0
_ERROR_UNKNOWN_ERROR = 1
_ERROR_UNEXPECTED_NODETYPE = 2
_ERROR_NOT_ENOUGH_ARGUMENTS_SPECIFIED = 3

_TIMEOUT_ERROR_EXIT = 124
_TIMEOUT_ERROR_KILL = 137
_TIMEOUT_ERROR_UNKNOWN = -9
_TIMEOUT_ERRORS = [
    _TIMEOUT_ERROR_EXIT, _TIMEOUT_ERROR_KILL, _TIMEOUT_ERROR_UNKNOWN]


def _get_node_type():
    result = _NODETYPE_UNKNOWN
    try:
        with open("/etc/nsx_issue") as fo:
            lines = fo.readlines()
        for line in lines:
            parts = line.strip().split(":")
            if len(parts) == 2 and parts[0] == "node-type":
                return parts[1].strip()
    except Exception:
        lg.warning(
            "Unable to determine node type\n%s" % traceback.format_exc())
        sys.stderr.write(
            "Unable to determine node type\n%s\n" % traceback.format_exc())
    return result


def main(argv):
    # Verify destination directory specified in argv
    if len(argv) < 4:
        lg.warning("Not enough arguments specified")
        sys.stderr.write("Not enough arguments specified\n")
        return _ERROR_NOT_ENOUGH_ARGUMENTS_SPECIFIED, None
    result = _OK
    node_id = argv[1]
    log_timestamp = argv[2]
    log_type = argv[3]
    dst_log_path = None
    dst_dir = "/tmp"

    node_type = _get_node_type()
    if _NODETYPE_MANAGER not in node_type and\
            _NODETYPE_CONTROLLER not in node_type and\
            _NODETYPE_EDGE not in node_type and\
            _NODETYPE_PUBLIC_CLOUD_GATEWAY not in node_type:
        return _ERROR_UNEXPECTED_NODETYPE, None
    uid = pwd.getpwnam(_PLATFORM_USR).pw_uid
    gid = grp.getgrnam(_PLATFORM_GRP).gr_gid
    cmd = [
        "/usr/bin/timeout", "--kill-after=1", str(_LOG_TIMEOUT),
        "/usr/bin/python", "-OO", _NAPI_AUDIT_LOG_SCRIPT,
        dst_dir, str(uid), str(gid), log_timestamp, log_type]
    try:
        env = {"PYTHONPATH": _PYTHONPATH}
        result, out, err = invoke_command(cmd, env=env)
        if result == 0:
            out = out.strip()
            tmp_path = os.path.join(dst_dir, os.path.basename(out))
            shutil.move(out, tmp_path)
            dst_log_path = tmp_path
        else:
            # Map timeout result from /usr/bin/timeout to 124. It can be:
            # * 124 or 137 as discussed in man timeout(1)
            # * -9 due to exit via signal (presumably when /usr/bin/timeout
            #   kills python -OO _NAPI_AUDIT_LOG_SCRIPT ...)
            if result in _TIMEOUT_ERRORS:
                result = _TIMEOUT_ERROR_EXIT
                lg.warning("Audit log generation timed out")
                sys.stderr.write("Audit log generation timed out\n")
            else:
                lg.warning(
                    "Unable to collect audit log: result: %s, "
                    "out: %s, err: %s" % (result, out, err))
                sys.stderr.write(
                    "Unable to collect audit log: result: %s, "
                    "out: %s, err: %s\n" % (result, out, err))
    except Exception:
        result = _ERROR_UNKNOWN_ERROR
        lg.warning(
            "Failed to generate or copy audit log to "
            "destination directory\n%s" % traceback.format_exc())
        sys.stderr.write(
            "Failed to generate or copy audit log to "
            "destination directory\n%s\n" % traceback.format_exc())

    return result, dst_log_path


if __name__ == "__main__":
    global lg
    nsx_logging.basicConfig(syslog=True, subcomp="audit-collect")
    lg = nsx_logging.getLogger(__name__)
    lg.setLevel(nsx_logging.INFO)
    lg.info("Start collecting audit logs with args %s", sys.argv[1:])
    result, dst_log_path = main(sys.argv)
    if result == _OK:
        sys.stdout.write(dst_log_path)
    sys.exit(result)
